"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.scheduleMissingTxsDataQuery = exports.cleanOldOriginalTxsMap = exports.cleanOldReceiptsMap = exports.queryTxDataFromArchivers = exports.collectMissingReceipts = exports.collectMissingTxDataFromArchivers = exports.processGossipData = exports.validateGossipData = exports.validateOriginalTxDataSchema = exports.storeOriginalTxData = exports.storeAccountData = exports.storeCycleData = exports.validateCycleData = exports.storeReceiptData = exports.verifyArchiverReceipt = exports.verifyReceiptData = exports.validateReceiptType = exports.storingAccountData = void 0;
const util_1 = require("util");
const Account = require("../dbstore/accounts");
const Transaction = require("../dbstore/transactions");
const Receipt = require("../dbstore/receipts");
const OriginalTxsData = require("../dbstore/originalTxsData");
const ProcessedTransaction = require("../dbstore/processedTxs");
const Crypto = require("../Crypto");
const Data_1 = require("./Data");
const Config_1 = require("../Config");
const Logger = require("../Logger");
const nestedCounters_1 = require("../profiler/nestedCounters");
const profiler_1 = require("../profiler/profiler");
const Cycles_1 = require("./Cycles");
const cycles_1 = require("../dbstore/cycles");
const State = require("../State");
const Utils = require("../Utils");
const GossipData_1 = require("./GossipData");
const P2P_1 = require("../P2P");
const GlobalAccount_1 = require("../GlobalAccount");
const DataLogWriter_1 = require("../Data/DataLogWriter");
const ShardFunctions_1 = require("../ShardFunctions");
const calculateAccountHash_1 = require("../shardeum/calculateAccountHash");
const verifyAppReceiptData_1 = require("../shardeum/verifyAppReceiptData");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const Helpers_1 = require("../types/ajv/Helpers");
const AJVSchemaEnum_1 = require("../types/enum/AJVSchemaEnum");
const transactionVerification_1 = require("../services/transactionVerification");
exports.storingAccountData = false;
const processedReceiptsMap = new Map();
const receiptsInValidationMap = new Map();
const processedOriginalTxsMap = new Map();
const originalTxsInValidationMap = new Map();
const missingReceiptsMap = new Map();
const missingOriginalTxsMap = new Map();
const collectingMissingReceiptsMap = new Map();
const collectingMissingOriginalTxsMap = new Map();
const verifyReceiptMajority = async (receipt, executionGroupNodes, minConfirmations = Config_1.config.RECEIPT_CONFIRMATIONS) => {
    /**
     * Note:
     * Currently, only the non-global receipt flow is implemented in `verifyReceiptMajority`,
     * `verifyReceiptOffline`, and `verifyNonGlobalTxReceiptWithValidators`. In the future,
     * global receipt methods can be added to maintain consistency. As of now, only offline
     * verification for global receipts is available, and `verifyGlobalTxReceiptWithValidators`
     * is not yet implemented.
     */
    // If robustQuery is disabled, do offline verification
    if (!Config_1.config.useRobustQueryForReceipt) {
        return verifyReceiptOffline(receipt, executionGroupNodes, minConfirmations);
    }
    return verifyReceiptWithValidators(receipt, executionGroupNodes, minConfirmations);
};
// Offline receipt verification
/**
 * Note:
 * The `verifyReceiptOffline` function is responsible for verifying receipts
 * without querying external validators, which is useful when the robust query
 * feature is disabled. It currently supports only non-global receipt verification.
 * Future enhancements should include validation logic for global receipts to ensure
 * comprehensive receipt verification. This will help maintain consistency and reliability
 * across all types of receipts processed by the system.
 *
 * The function delegates the verification process to `verifyNonGlobalTxReceiptOffline`
 * based on the whether the receipt is global or not.
 */
const verifyReceiptOffline = async (receipt, executionGroupNodes, minConfirmations) => {
    return verifyNonGlobalTxReceiptOffline(receipt, executionGroupNodes, minConfirmations);
};
/**
 * Note:
 * The `verifyReceiptWithValidators` function currently supports only
 * non-global receipt verification. Future enhancements should include
 * validation logic for global receipts to ensure comprehensive receipt
 * verification. This will help in maintaining consistency and reliability
 * across all types of receipts processed by the system.
 */
const verifyReceiptWithValidators = async (receipt, executionGroupNodes, minConfirmations = Config_1.config.RECEIPT_CONFIRMATIONS) => {
    return verifyNonGlobalTxReceiptWithValidators(receipt, executionGroupNodes, minConfirmations);
};
/**
 * Calls the /get-tx-receipt endpoint of the nodes in the execution group of the receipt to verify the receipt. If "RECEIPT_CONFIRMATIONS" number of nodes return the same receipt, the receipt is deemed valid.
 * @param receipt
 * @param executionGroupNodes
 * @param minConfirmations
 * @returns boolean
 */
const verifyNonGlobalTxReceiptWithValidators = async (receipt, executionGroupNodes, minConfirmations = Config_1.config.RECEIPT_CONFIRMATIONS) => {
    const result = { success: false };
    // Created signedData with full_receipt = false outside of queryReceipt to avoid signing the same data multiple times
    let signedData = Crypto.sign({
        txId: receipt.tx.txId,
        timestamp: receipt.tx.timestamp,
        full_receipt: false,
    });
    const queryReceipt = async (node) => {
        const QUERY_RECEIPT_TIMEOUT_SECOND = 2;
        try {
            return (await (0, P2P_1.postJson)(`http://${node.ip}:${node.port}/get-tx-receipt`, signedData, QUERY_RECEIPT_TIMEOUT_SECOND));
        }
        catch (error) {
            Logger.mainLogger.error('Error in /get-tx-receipt:', error);
            if (nestedCounters_1.nestedCountersInstance)
                nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Error_in_get-tx-receipt');
            return null;
        }
    };
    // Use only random 5 x Receipt Confirmations number of nodes from the execution group to reduce the number of nodes to query in large execution groups
    const filteredExecutionGroupNodes = Utils.getRandomItemFromArr(executionGroupNodes, 0, 5 * Config_1.config.RECEIPT_CONFIRMATIONS);
    const isReceiptEqual = (receipt1, receipt2) => {
        if (!receipt1 || !receipt2)
            return false;
        const r1 = Utils.deepCopy(receipt1);
        const r2 = Utils.deepCopy(receipt2);
        // The confirmOrChallenge node could be different in the two receipts, so we need to remove it before comparing
        delete r1?.confirmOrChallenge?.nodeId;
        delete r1?.confirmOrChallenge?.sign;
        delete r2?.confirmOrChallenge?.nodeId;
        delete r2?.confirmOrChallenge?.sign;
        const equivalent = (0, util_1.isDeepStrictEqual)(r1, r2);
        return equivalent;
    };
    const robustQuery = await Utils.robustQuery(filteredExecutionGroupNodes, (execNode) => queryReceipt(execNode), (rec1, rec2) => isReceiptEqual(rec1?.receipt, rec2?.receipt), minConfirmations, false, // set shuffleNodes to false,
    500, // Add 500 ms delay
    true);
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('robustQuery', receipt.tx.txId, robustQuery);
    if (!robustQuery || !robustQuery.value || !robustQuery.value.receipt) {
        Logger.mainLogger.error(`❌ 'null' response from all nodes in receipt-validation for txId: ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}
      }`);
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'null_response_from_all_nodes_in_receipt-validation');
        return result;
    }
    const robustQueryReceipt = robustQuery.value.receipt;
    if (robustQuery.count < minConfirmations) {
        // Wait for 500ms and try fetching the receipt from the nodes that did not respond in the robustQuery
        await Utils.sleep(500);
        let requiredConfirmations = minConfirmations - robustQuery.count;
        let nodesToQuery = executionGroupNodes.filter((node) => !robustQuery.nodes.some((n) => n.publicKey === node.publicKey));
        let retryCount = 5; // Retry 5 times
        while (requiredConfirmations > 0) {
            if (nodesToQuery.length === 0) {
                if (retryCount === 0)
                    break;
                // Wait for 500ms and try again
                await Utils.sleep(500);
                nodesToQuery = executionGroupNodes.filter((node) => !robustQuery.nodes.some((n) => n.publicKey === node.publicKey));
                retryCount--;
                if (nodesToQuery.length === 0)
                    break;
            }
            const node = nodesToQuery[0];
            nodesToQuery.splice(0, 1);
            const receiptResult = await queryReceipt(node);
            if (!receiptResult || !receiptResult.receipt)
                continue;
            if (isReceiptEqual(robustQueryReceipt, receiptResult.receipt)) {
                requiredConfirmations--;
                robustQuery.nodes.push(node);
            }
        }
    }
    // Check if the robustQueryReceipt is the same as our receipt
    const sameReceipt = isReceiptEqual(receipt.signedReceipt, robustQueryReceipt);
    if (!sameReceipt) {
        Logger.mainLogger.debug(`Found different receipt in robustQuery ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Found_different_receipt_in_robustQuery');
        if (Config_1.config.VERBOSE)
            Logger.mainLogger.debug(receipt.signedReceipt);
        if (Config_1.config.VERBOSE)
            Logger.mainLogger.debug(robustQueryReceipt);
        // update signedData with full_receipt = true
        signedData = Crypto.sign({ txId: receipt.tx.txId, timestamp: receipt.tx.timestamp, full_receipt: true });
        for (const node of robustQuery.nodes) {
            const fullReceiptResult = await queryReceipt(node);
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug(`'fullReceiptResult ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`, fullReceiptResult);
            if (!fullReceiptResult || !fullReceiptResult.receipt) {
                Logger.mainLogger.error(`Got fullReceiptResult null from robustQuery node for ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
                continue;
            }
            const fullReceipt = fullReceiptResult.receipt;
            if (isReceiptEqual(fullReceipt.signedReceipt, robustQueryReceipt) &&
                (0, exports.validateReceiptType)(fullReceipt)) {
                if (Config_1.config.verifyAppReceiptData) {
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionStart('Verify_app_receipt_data');
                    if (nestedCounters_1.nestedCountersInstance)
                        nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Verify_app_receipt_data');
                    const { valid, needToSave } = await (0, verifyAppReceiptData_1.verifyAppReceiptData)(receipt);
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionEnd('Verify_app_receipt_data');
                    if (!valid) {
                        Logger.mainLogger.error(`The app receipt verification failed from robustQuery nodes ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
                        continue;
                    }
                    if (!needToSave) {
                        Logger.mainLogger.debug(`Found valid full receipt in robustQuery ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
                        Logger.mainLogger.error(`Found valid receipt from robustQuery: but no need to save ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
                        return { success: false };
                    }
                }
                if (Config_1.config.verifyAccountData && receipt.globalModification === false) {
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionStart('Verify_receipt_account_data');
                    if (nestedCounters_1.nestedCountersInstance)
                        nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Verify_receipt_account_data');
                    const result = (0, calculateAccountHash_1.verifyAccountHash)(fullReceipt);
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionEnd('Verify_receipt_account_data');
                    if (!result) {
                        Logger.mainLogger.error(`The account verification failed from robustQuery nodes ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
                        continue;
                    }
                }
                if (Config_1.config.verifyReceiptData) {
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionStart('Verify_app_receipt_data');
                    if (nestedCounters_1.nestedCountersInstance)
                        nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Verify_app_receipt_data');
                    const { success } = await (0, exports.verifyReceiptData)(fullReceipt, false);
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionEnd('Verify_app_receipt_data');
                    if (!success) {
                        Logger.mainLogger.error(`The receipt validation failed from robustQuery nodes ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
                        continue;
                    }
                }
                Logger.mainLogger.debug(`Found valid full receipt in robustQuery ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
                if (nestedCounters_1.nestedCountersInstance)
                    nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Found_valid_full_receipt_in_robustQuery');
                return { success: true, newReceipt: fullReceipt };
            }
            else {
                Logger.mainLogger.error(`The receipt validation failed from robustQuery nodes ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
                Logger.mainLogger.error(lib_types_1.Utils.safeStringify(robustQueryReceipt), lib_types_1.Utils.safeStringify(fullReceipt));
            }
        }
        Logger.mainLogger.error(`No valid full receipt found in robustQuery ${receipt.tx.txId} , ${receipt.cycle}, ${receipt.tx.timestamp}`);
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'No_valid_full_receipt_found_in_robustQuery');
        return { success: false };
    }
    return { success: true };
};
// Offline global receipt verification
const verifyGlobalTxreceiptOffline = async (receipt) => {
    const appliedReceipt = receipt.signedReceipt;
    const result = { success: false };
    const { txId, timestamp } = receipt.tx;
    const { executionShardKey, cycle } = receipt;
    const cycleShardData = Cycles_1.shardValuesByCycle.get(cycle);
    const { homePartition } = ShardFunctions_1.default.addressToPartition(cycleShardData.shardGlobals, executionShardKey);
    const { signs } = appliedReceipt;
    // Refer to https://github.com/shardeum/shardus-core/blob/7d8877b7e1a5b18140f898a64b932182d8a35298/src/p2p/GlobalAccounts.ts#L397
    let votingGroupCount = cycleShardData.shardGlobals.nodesPerConsenusGroup;
    if (votingGroupCount > cycleShardData.nodes.length) {
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'votingGroupCount_greater_than_nodes_length');
        Logger.mainLogger.error('votingGroupCount_greater_than_nodes_length', votingGroupCount, cycleShardData.nodes.length);
        votingGroupCount = cycleShardData.nodes.length;
    }
    let isReceiptMajority = (signs.length / votingGroupCount) * 100 >= Config_1.config.requiredMajorityVotesPercentage;
    if (!isReceiptMajority) {
        Logger.mainLogger.error(`Invalid receipt globalModification signs count is less than ${Config_1.config.requiredMajorityVotesPercentage}% of the votingGroupCount, ${signs.length}, ${votingGroupCount}`);
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', `Invalid_receipt_globalModification_signs_count_less_than_${Config_1.config.requiredMajorityVotesPercentage}%`);
        return result;
    }
    const nodeMap = new Map();
    // Fill the map with nodes keyed by their public keys
    cycleShardData.nodes.forEach((node) => {
        if (node.publicKey) {
            nodeMap.set(node.publicKey, node);
        }
    });
    // Using a set to store the unique signers to avoid duplicates
    const uniqueSigners = new Set();
    for (const sign of signs) {
        const { owner: nodePubKey } = sign;
        // Get the node id from the public key
        const node = nodeMap.get(nodePubKey);
        if (node == null) {
            Logger.mainLogger.error(`The node with public key ${nodePubKey} of the receipt ${txId} with ${timestamp} is not in the active nodesList of cycle ${cycle}`);
            if (nestedCounters_1.nestedCountersInstance)
                nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'globalModification_sign_owner_not_in_active_nodesList');
            continue;
        }
        // Check if the node is in the execution group
        if (!cycleShardData.parititionShardDataMap.get(homePartition).coveredBy[node.id]) {
            Logger.mainLogger.error(`The node with public key ${nodePubKey} of the receipt ${txId} with ${timestamp} is not in the execution group of the tx`);
            if (nestedCounters_1.nestedCountersInstance)
                nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'globalModification_sign_node_not_in_execution_group_of_tx');
            continue;
        }
        uniqueSigners.add(nodePubKey);
    }
    isReceiptMajority = (uniqueSigners.size / votingGroupCount) * 100 >= Config_1.config.requiredMajorityVotesPercentage;
    if (!isReceiptMajority) {
        Logger.mainLogger.error(`Invalid receipt globalModification valid signs count is less than votingGroupCount ${uniqueSigners.size}, ${votingGroupCount}`);
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Invalid_receipt_globalModification_valid_signs_count_less_than_votingGroupCount');
        return result;
    }
    const requiredSignatures = Math.floor(votingGroupCount * (Config_1.config.requiredMajorityVotesPercentage / 100));
    return { success: true, requiredSignatures };
};
const verifyNonGlobalTxReceiptOffline = async (receipt, executionGroupNodes, minConfirmations) => {
    // Code for normal receipts verification
    const normalReceipt = receipt.signedReceipt;
    const validSigners = new Set();
    if (!normalReceipt.signaturePack || !Array.isArray(normalReceipt.signaturePack)) {
        return { success: false };
    }
    for (const signature of normalReceipt.signaturePack) {
        if (!signature || !signature.owner)
            continue;
        const node = executionGroupNodes.find((n) => n.publicKey.toLowerCase() === signature.owner.toLowerCase());
        if (!node)
            continue;
        try {
            const voteHash = calculateVoteHash(normalReceipt.proposal);
            const appliedVoteHash = {
                txid: receipt.tx.txId,
                voteHash
            };
            if (Crypto.verify({ ...appliedVoteHash, sign: signature })) {
                validSigners.add(signature.owner);
            }
        }
        catch (error) {
            console.error('Error verifying signature:', error);
            continue;
        }
    }
    return {
        success: validSigners.size >= minConfirmations
    };
};
/**
 * Validate type and field existence of the receipt data before processing it further
 * @param receipt
 * @returns boolean
 */
const validateReceiptType = (receipt) => {
    // Validate against the Receipt schema will come when archiver is syncing from another archiver
    const errors_validation_receipt = (0, Helpers_1.verifyPayload)(AJVSchemaEnum_1.AJVSchemaEnum.Receipt, receipt);
    if (!errors_validation_receipt) {
        return true; // Valid Receipt
    }
    // Validate against the ArchiverReceipt schema this will be used when receipt object is getting received from validator
    const errors_validation_archiver_receipt = (0, Helpers_1.verifyPayload)(AJVSchemaEnum_1.AJVSchemaEnum.ArchiverReceipt, receipt);
    if (!errors_validation_archiver_receipt) {
        return true; // Valid ArchiverReceipt
    }
    // If neither validation passes, log the errors and return false
    Logger.mainLogger.error('Invalid Receipt', {
        receiptType: errors_validation_receipt ? 'ArchiverReceipt' : 'Receipt',
        receiptErrors: errors_validation_receipt || errors_validation_archiver_receipt,
    }, 'where receipt was', lib_types_1.Utils.safeStringify(receipt));
    return false; // Invalid receipt
};
exports.validateReceiptType = validateReceiptType;
const verifyReceiptData = async (receipt, checkReceiptRobust = true) => {
    const result = { success: false };
    // Check the signed nodes are part of the execution group nodes of the tx
    const { executionShardKey, cycle, globalModification } = receipt;
    const { txId, timestamp } = receipt.tx;
    if (Config_1.config.VERBOSE) {
        const currentTimestamp = Date.now();
        // Console log the timetaken between the receipt timestamp and the current time ( both in ms and s)
        console.log(`Time taken between receipt timestamp and current time: ${txId}`, `${currentTimestamp - timestamp} ms`, `${(currentTimestamp - timestamp) / 1000} s`);
    }
    const currentCycle = (0, Cycles_1.getCurrentCycleCounter)();
    if (currentCycle - cycle > 2) {
        Logger.mainLogger.error(`Found receipt with cycle older than 2 cycles ${txId}, ${cycle}, ${timestamp}, ${currentCycle}`);
    }
    const cycleShardData = Cycles_1.shardValuesByCycle.get(cycle);
    if (!cycleShardData) {
        Logger.mainLogger.error('Cycle shard data not found');
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Cycle_shard_data_not_found');
        return result;
    }
    // Determine the home partition index of the primary account (executionShardKey)
    const { homePartition } = ShardFunctions_1.default.addressToPartition(cycleShardData.shardGlobals, executionShardKey);
    let globalReceiptValidationErrors;
    try {
        // Validate if receipt is a global modification receipt using AJV
        globalReceiptValidationErrors = (0, Helpers_1.verifyPayload)(AJVSchemaEnum_1.AJVSchemaEnum.GlobalTxReceipt, receipt?.signedReceipt);
    }
    catch (error) {
        globalReceiptValidationErrors = true;
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', `Failed to validate receipt schema txId: ${txId}, cycle: ${cycle}, timestamp: ${timestamp}, error: ${error}`);
        Logger.mainLogger.error(`Failed to validate receipt schema txId: ${txId}, cycle: ${cycle}, timestamp: ${timestamp}, error: ${error}`);
        return result;
    }
    // If the receipt is a global modification receipt, validate the receipt
    if (!globalReceiptValidationErrors) {
        return verifyGlobalTxreceiptOffline(receipt);
    }
    const { signaturePack } = receipt.signedReceipt;
    if (Config_1.config.newPOQReceipt === false) {
        // Refer to https://github.com/shardeum/shardus-core/blob/f7000c36faa0cd1e0832aa1e5e3b1414d32dcf66/src/state-manager/TransactionConsensus.ts#L1406
        let votingGroupCount = cycleShardData.shardGlobals.nodesPerConsenusGroup;
        if (votingGroupCount > cycleShardData.nodes.length) {
            votingGroupCount = cycleShardData.nodes.length;
        }
        const requiredSignatures = Config_1.config.usePOQo === true
            ? Math.ceil(votingGroupCount * Config_1.config.requiredVotesPercentage)
            : Math.round(votingGroupCount * Config_1.config.requiredVotesPercentage);
        if (signaturePack.length < requiredSignatures) {
            Logger.mainLogger.error(`Invalid receipt appliedReceipt signatures count is less than requiredSignatures, ${signaturePack.length}, ${requiredSignatures}`);
            if (nestedCounters_1.nestedCountersInstance)
                nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Invalid_receipt_appliedReceipt_signatures_count_less_than_requiredSignatures');
            return result;
        }
        // Using a set to store the unique signatures to avoid duplicates
        const uniqueSigners = new Set();
        for (const signature of signaturePack) {
            const { owner: nodePubKey } = signature;
            // Get the node id from the public key
            const node = cycleShardData.nodes.find((node) => node.publicKey === nodePubKey);
            if (node == null) {
                Logger.mainLogger.error(`The node with public key ${nodePubKey} of the receipt ${txId}} with ${timestamp} is not in the active nodesList of cycle ${cycle}`);
                if (nestedCounters_1.nestedCountersInstance)
                    nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'appliedReceipt_signature_owner_not_in_active_nodesList');
                continue;
            }
            // Check if the node is in the execution group
            if (!cycleShardData.parititionShardDataMap.get(homePartition).coveredBy[node.id]) {
                Logger.mainLogger.error(`The node with public key ${nodePubKey} of the receipt ${txId} with ${timestamp} is not in the execution group of the tx`);
                if (nestedCounters_1.nestedCountersInstance)
                    nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'appliedReceipt_signature_node_not_in_execution_group_of_tx');
                continue;
            }
            uniqueSigners.add(nodePubKey);
        }
        if (uniqueSigners.size < requiredSignatures) {
            Logger.mainLogger.error(`Invalid receipt appliedReceipt valid signatures count is less than requiredSignatures ${uniqueSigners.size}, ${requiredSignatures}`);
            if (nestedCounters_1.nestedCountersInstance)
                nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Invalid_receipt_appliedReceipt_valid_signatures_count_less_than_requiredSignatures');
            return result;
        }
        return { success: true, requiredSignatures };
    }
    // const { confirmOrChallenge } = appliedReceipt as Receipt.AppliedReceipt2
    // // Check if the appliedVote node is in the execution group
    // if (!cycleShardData.nodeShardDataMap.has(appliedVote.node_id)) {
    //   Logger.mainLogger.error('Invalid receipt appliedReceipt appliedVote node is not in the active nodesList')
    //   if (nestedCountersInstance)
    //     nestedCountersInstance.countEvent('receipt', 'Invalid_receipt_appliedVote_node_not_in_active_nodesList')
    //   return result
    // }
    // if (appliedVote.sign.owner !== cycleShardData.nodeShardDataMap.get(appliedVote.node_id).node.publicKey) {
    //   Logger.mainLogger.error(
    //     'Invalid receipt appliedReceipt appliedVote node signature owner and node public key does not match'
    //   )
    //   if (nestedCountersInstance)
    //     nestedCountersInstance.countEvent(
    //       'receipt',
    //       'Invalid_receipt_appliedVote_node_signature_owner_and_node_public_key_does_not_match'
    //     )
    //   return result
    // }
    // if (!cycleShardData.parititionShardDataMap.get(homePartition).coveredBy[appliedVote.node_id]) {
    //   Logger.mainLogger.error(
    //     'Invalid receipt appliedReceipt appliedVote node is not in the execution group of the tx'
    //   )
    //   if (nestedCountersInstance)
    //     nestedCountersInstance.countEvent(
    //       'receipt',
    //       'Invalid_receipt_appliedVote_node_not_in_execution_group_of_tx'
    //     )
    //   return result
    // }
    // if (!Crypto.verify(appliedVote)) {
    //   Logger.mainLogger.error('Invalid receipt appliedReceipt appliedVote signature verification failed')
    //   if (nestedCountersInstance)
    //     nestedCountersInstance.countEvent(
    //       'receipt',
    //       'Invalid_receipt_appliedVote_signature_verification_failed'
    //     )
    //   return result
    // }
    // // Check if the confirmOrChallenge node is in the execution group
    // if (!cycleShardData.nodeShardDataMap.has(confirmOrChallenge.nodeId)) {
    //   Logger.mainLogger.error(
    //     'Invalid receipt appliedReceipt confirmOrChallenge node is not in the active nodesList'
    //   )
    //   if (nestedCountersInstance)
    //     nestedCountersInstance.countEvent(
    //       'receipt',
    //       'Invalid_receipt_confirmOrChallenge_node_not_in_active_nodesList'
    //     )
    //   return result
    // }
    // if (
    //   confirmOrChallenge.sign.owner !==
    //   cycleShardData.nodeShardDataMap.get(confirmOrChallenge.nodeId).node.publicKey
    // ) {
    //   Logger.mainLogger.error(
    //     'Invalid receipt appliedReceipt confirmOrChallenge node signature owner and node public key does not match'
    //   )
    //   if (nestedCountersInstance)
    //     nestedCountersInstance.countEvent(
    //       'receipt',
    //       'Invalid_receipt_confirmOrChallenge_signature_owner_and_node_public_key_does_not_match'
    //     )
    //   return result
    // }
    // if (!cycleShardData.parititionShardDataMap.get(homePartition).coveredBy[confirmOrChallenge.nodeId]) {
    //   Logger.mainLogger.error(
    //     'Invalid receipt appliedReceipt confirmOrChallenge node is not in the execution group of the tx'
    //   )
    //   if (nestedCountersInstance)
    //     nestedCountersInstance.countEvent(
    //       'receipt',
    //       'Invalid_receipt_confirmOrChallenge_node_not_in_execution_group_of_tx'
    //     )
    //   return result
    // }
    // if (!Crypto.verify(confirmOrChallenge)) {
    //   Logger.mainLogger.error('Invalid receipt appliedReceipt confirmOrChallenge signature verification failed')
    //   if (nestedCountersInstance)
    //     nestedCountersInstance.countEvent(
    //       'receipt',
    //       'Invalid_receipt_confirmOrChallenge_signature_verification_failed'
    //     )
    //   return result
    // }
    if (!checkReceiptRobust)
        return { success: true };
    // List the execution group nodes of the tx, Use them to robustQuery to verify the receipt
    const executionGroupNodes = Object.values(cycleShardData.parititionShardDataMap.get(homePartition).coveredBy);
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('executionGroupNodes', receipt.tx.txId, executionGroupNodes);
    const minConfirmations = Data_1.nodesPerConsensusGroup > Config_1.config.RECEIPT_CONFIRMATIONS
        ? Config_1.config.RECEIPT_CONFIRMATIONS
        : Math.ceil(Config_1.config.RECEIPT_CONFIRMATIONS / 2); // 3 out of 5 nodes
    const { success, newReceipt } = await verifyReceiptMajority(receipt, executionGroupNodes, minConfirmations);
    if (!success) {
        Logger.mainLogger.error('Invalid receipt: Robust check failed');
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Invalid_receipt_robust_check_failed');
        return result;
    }
    if (newReceipt)
        return { success: true, requiredSignatures: 0, newReceipt };
    return { success: true };
};
exports.verifyReceiptData = verifyReceiptData;
const verifyAppliedReceiptSignatures = (receipt, requiredSignatures, failedReasons = [], nestedCounterMessages = []) => {
    const result = { success: false, failedReasons, nestedCounterMessages };
    const { globalModification, cycle, executionShardKey } = receipt;
    const { txId: txid, timestamp } = receipt.tx;
    let globalReceiptValidationErrors; // This is used to store the validation errors of the globalTxReceipt
    try {
        globalReceiptValidationErrors = (0, Helpers_1.verifyPayload)(AJVSchemaEnum_1.AJVSchemaEnum.GlobalTxReceipt, receipt?.signedReceipt);
    }
    catch (error) {
        globalReceiptValidationErrors = true;
        failedReasons.push(`Invalid Global Tx Receipt error: ${error}. txId ${receipt.tx.txId} , cycle ${receipt.cycle} , timestamp ${receipt.tx.timestamp}`);
        nestedCounterMessages.push(`Invalid Global Tx Receipt error: ${error}. txId ${receipt.tx.txId} , cycle ${receipt.cycle} , timestamp ${receipt.tx.timestamp}`);
        return result;
    }
    // If the globalReceiptValidationErrors is null, then the receipt is a globalModification receipt
    if (!globalReceiptValidationErrors) {
        const appliedReceipt = receipt.signedReceipt;
        // Refer to https://github.com/shardeum/shardus-core/blob/7d8877b7e1a5b18140f898a64b932182d8a35298/src/p2p/GlobalAccounts.ts#L294
        const { signs, tx } = appliedReceipt;
        const cycleShardData = Cycles_1.shardValuesByCycle.get(cycle);
        const { homePartition } = ShardFunctions_1.default.addressToPartition(cycleShardData.shardGlobals, executionShardKey);
        const nodeMap = new Map();
        // Fill the map with nodes keyed by their public keys
        cycleShardData.nodes.forEach((node) => {
            if (node.publicKey) {
                nodeMap.set(node.publicKey, node);
            }
        });
        const acceptableSigners = new Set();
        for (const sign of signs) {
            const { owner: nodePubKey } = sign;
            // Get the node id from the public key
            const node = nodeMap.get(nodePubKey);
            if (node == null) {
                Logger.mainLogger.error(`The node with public key ${nodePubKey} of the receipt ${txid} with ${timestamp} is not in the active nodesList of cycle ${cycle}`);
                if (nestedCounters_1.nestedCountersInstance)
                    nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'globalModification_sign_owner_not_in_active_nodesList');
                continue;
            }
            // Check if the node is in the execution group
            if (!cycleShardData.parititionShardDataMap.get(homePartition).coveredBy[node.id]) {
                Logger.mainLogger.error(`The node with public key ${nodePubKey} of the receipt ${txid} with ${timestamp} is not in the execution group of the tx`);
                if (nestedCounters_1.nestedCountersInstance)
                    nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'globalModification_sign_node_not_in_execution_group_of_tx');
                continue;
            }
            acceptableSigners.add(sign);
        }
        // Using a map to store the good signatures to avoid duplicates
        const goodSignatures = new Map();
        for (const sign of acceptableSigners) {
            if (Crypto.verify({ ...tx, sign: sign })) {
                goodSignatures.set(sign.owner, sign);
                // Break the loop if the required number of good signatures are found
                if (goodSignatures.size >= requiredSignatures)
                    break;
            }
        }
        if (goodSignatures.size < requiredSignatures) {
            failedReasons.push(`Invalid receipt globalModification valid signs count is less than requiredSignatures ${txid}, ${goodSignatures.size}, ${requiredSignatures}`);
            nestedCounterMessages.push('Invalid_receipt_globalModification_valid_signs_count_less_than_requiredSignatures');
            return result;
        }
        return { success: true };
    }
    const { proposal, signaturePack, voteOffsets } = receipt.signedReceipt;
    // Refer to https://github.com/shardeum/shardus-core/blob/50b6d00f53a35996cd69210ea817bee068a893d6/src/state-manager/TransactionConsensus.ts#L2799
    const voteHash = calculateVoteHash(proposal, failedReasons, nestedCounterMessages);
    // Refer to https://github.com/shardeum/shardus-core/blob/50b6d00f53a35996cd69210ea817bee068a893d6/src/state-manager/TransactionConsensus.ts#L2663
    const appliedVoteHash = {
        txid,
        voteHash,
    };
    // Using a map to store the good signatures to avoid duplicates
    const goodSignatures = new Map();
    for (const [index, signature] of signaturePack.entries()) {
        if (Crypto.verify({ ...appliedVoteHash, sign: signature, voteTime: voteOffsets.at(index) })) {
            goodSignatures.set(signature.owner, signature);
            // Break the loop if the required number of good signatures are found
            if (goodSignatures.size >= requiredSignatures)
                break;
        }
        else {
            failedReasons.push(`Found invalid signature in receipt signedReceipt ${txid}, ${signature.owner}, ${index}`);
            nestedCounterMessages.push('Found_invalid_signature_in_receipt_signedReceipt');
        }
    }
    if (goodSignatures.size < requiredSignatures) {
        failedReasons.push(`Invalid receipt signedReceipt valid signatures count is less than requiredSignatures ${txid}, ${goodSignatures.size}, ${requiredSignatures}`);
        nestedCounterMessages.push('Invalid_receipt_signedReceipt_valid_signatures_count_less_than_requiredSignatures');
        return result;
    }
    return { success: true };
};
const calculateVoteHash = (vote, failedReasons = [], nestedCounterMessages = []) => {
    try {
        if (Config_1.config.usePOQo === true && vote.applied !== undefined) {
            const proposal = vote;
            const applyStatus = {
                applied: proposal.applied,
                cantApply: proposal.cant_preApply,
            };
            const accountsHash = Crypto.hash(Crypto.hashObj(proposal.accountIDs) +
                Crypto.hashObj(proposal.beforeStateHashes) +
                Crypto.hashObj(proposal.afterStateHashes));
            const proposalHash = Crypto.hash(Crypto.hashObj(applyStatus) + accountsHash + proposal.appReceiptDataHash);
            return proposalHash;
        }
        else if (Config_1.config.usePOQo === true) {
            const appliedVote = vote;
            const appliedHash = {
                applied: appliedVote.transaction_result,
                cantApply: appliedVote.cant_apply,
            };
            const stateHash = {
                account_id: appliedVote.account_id,
                account_state_hash_after: appliedVote.account_state_hash_after,
                account_state_hash_before: appliedVote.account_state_hash_before,
            };
            const appDataHash = {
                app_data_hash: appliedVote.app_data_hash,
            };
            const voteToHash = {
                appliedHash: Crypto.hashObj(appliedHash),
                stateHash: Crypto.hashObj(stateHash),
                appDataHash: Crypto.hashObj(appDataHash),
            };
            return Crypto.hashObj(voteToHash);
        }
        return Crypto.hashObj({ ...vote, node_id: '' });
    }
    catch {
        failedReasons.push('Error in calculateVoteHash', vote);
        nestedCounterMessages.push('Error_in_calculateVoteHash');
        return '';
    }
};
const verifyArchiverReceipt = async (receipt, requiredSignatures) => {
    const { txId, timestamp } = receipt.tx;
    const existingReceipt = await Receipt.queryReceiptByReceiptId(txId);
    const failedReasons = [];
    const nestedCounterMessages = [];
    if (Config_1.config.verifyAppReceiptData) {
        // if (profilerInstance) profilerInstance.profileSectionStart('Verify_app_receipt_data')
        // if (nestedCountersInstance) nestedCountersInstance.countEvent('receipt', 'Verify_app_receipt_data')
        const { valid, needToSave } = await (0, verifyAppReceiptData_1.verifyAppReceiptData)(receipt, existingReceipt, failedReasons, nestedCounterMessages);
        // if (profilerInstance) profilerInstance.profileSectionEnd('Verify_app_receipt_data')
        if (!valid) {
            failedReasons.push(`Invalid receipt: App Receipt Verification failed ${txId}, ${receipt.cycle}, ${timestamp}`);
            nestedCounterMessages.push('Invalid_receipt_app_receipt_verification_failed');
            return { success: false, failedReasons, nestedCounterMessages };
        }
        if (!needToSave) {
            failedReasons.push(`Valid receipt: but no need to save ${txId}, ${receipt.cycle}, ${timestamp}`);
            nestedCounterMessages.push('Valid_receipt_but_no_need_to_save');
            return { success: false, failedReasons, nestedCounterMessages };
        }
    }
    if (Config_1.config.verifyAccountData) {
        // if (profilerInstance) profilerInstance.profileSectionStart('Verify_receipt_account_data')
        // if (nestedCountersInstance) nestedCountersInstance.countEvent('receipt', 'Verify_receipt_account_data')
        const result = (0, calculateAccountHash_1.verifyAccountHash)(receipt, failedReasons, nestedCounterMessages);
        // if (profilerInstance) profilerInstance.profileSectionEnd('Verify_receipt_account_data')
        if (!result) {
            failedReasons.push(`Invalid receipt: Account Verification failed ${txId}, ${receipt.cycle}, ${timestamp}`);
            nestedCounterMessages.push('Invalid_receipt_account_verification_failed');
            return { success: false, failedReasons, nestedCounterMessages };
        }
    }
    if (Config_1.config.verifyReceiptSignaturesSeparately) {
        // if (profilerInstance) profilerInstance.profileSectionStart('Verify_receipt_signatures_data')
        // if (nestedCountersInstance) nestedCountersInstance.countEvent('receipt', 'Verify_receipt_signatures_data')
        const { success } = verifyAppliedReceiptSignatures(receipt, requiredSignatures, failedReasons, nestedCounterMessages);
        // if (profilerInstance) profilerInstance.profileSectionEnd('Verify_receipt_signatures_data')
        if (!success) {
            failedReasons.push(`Invalid receipt: Verification failed ${txId}, ${receipt.cycle}, ${timestamp}`);
            nestedCounterMessages.push('Invalid_receipt_verification_failed');
            return { success: false, failedReasons, nestedCounterMessages };
        }
    }
    return { success: true, failedReasons, nestedCounterMessages };
};
exports.verifyArchiverReceipt = verifyArchiverReceipt;
const storeReceiptData = async (receipts, senderInfo = '', verifyData = false, saveOnlyGossipData = false) => {
    if (!receipts || !Array.isArray(receipts) || receipts.length <= 0)
        return;
    const bucketSize = 1000;
    let combineReceipts = [];
    let combineAccounts = [];
    let combineTransactions = [];
    let combineProcessedTxs = [];
    let txDataList = [];
    if (saveOnlyGossipData)
        return;
    for (let receipt of receipts) {
        const txId = receipt?.tx?.txId;
        const timestamp = receipt?.tx?.timestamp;
        if (!txId || !timestamp)
            continue;
        if ((processedReceiptsMap.has(txId) && processedReceiptsMap.get(txId) === timestamp) ||
            (receiptsInValidationMap.has(txId) && receiptsInValidationMap.get(txId) === timestamp)) {
            if (Config_1.config.VERBOSE)
                console.log('RECEIPT', 'Skip', txId, timestamp, senderInfo);
            continue;
        }
        if (Config_1.config.VERBOSE)
            console.log('RECEIPT', 'Validate', txId, timestamp, senderInfo);
        receiptsInValidationMap.set(txId, timestamp);
        if (profiler_1.profilerInstance)
            profiler_1.profilerInstance.profileSectionStart('Validate_receipt');
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Validate_receipt');
        if (!(0, exports.validateReceiptType)(receipt)) {
            Logger.mainLogger.error('Invalid receipt: Validation failed', txId, receipt.cycle, timestamp);
            receiptsInValidationMap.delete(txId);
            if (nestedCounters_1.nestedCountersInstance)
                nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Invalid_receipt_validation_failed');
            if (profiler_1.profilerInstance)
                profiler_1.profilerInstance.profileSectionEnd('Validate_receipt');
            continue;
        }
        if (verifyData) {
            // if (config.usePOQo === false) {
            // const existingReceipt = await Receipt.queryReceiptByReceiptId(txId)
            // if (
            //   existingReceipt &&
            //   receipt.appliedReceipt &&
            //   receipt.appliedReceipt.confirmOrChallenge &&
            //   receipt.appliedReceipt.confirmOrChallenge.message === 'challenge'
            // ) {
            //   // If the existing receipt is confirmed, and the new receipt is challenged, then skip saving the new receipt
            //   if (existingReceipt.appliedReceipt.confirmOrChallenge.message === 'confirm') {
            //     Logger.mainLogger.error(
            //       `Existing receipt is confirmed, but new receipt is challenged ${txId}, ${receipt.cycle}, ${timestamp}`
            //     )
            //     receiptsInValidationMap.delete(txId)
            //     if (nestedCountersInstance)
            //       nestedCountersInstance.countEvent(
            //         'receipt',
            //         'Existing_receipt_is_confirmed_but_new_receipt_is_challenged'
            //       )
            //     if (profilerInstance) profilerInstance.profileSectionEnd('Validate_receipt')
            //     continue
            //   }
            // }
            // }
            if (Config_1.config.verifyReceiptData) {
                const { success, requiredSignatures, newReceipt } = await (0, exports.verifyReceiptData)(receipt);
                if (!success) {
                    Logger.mainLogger.error('Invalid receipt: Verification failed', txId, receipt.cycle, timestamp);
                    receiptsInValidationMap.delete(txId);
                    if (nestedCounters_1.nestedCountersInstance)
                        nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Invalid_receipt_verification_failed');
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionEnd('Validate_receipt');
                    continue;
                }
                if (newReceipt)
                    receipt = newReceipt;
                if (profiler_1.profilerInstance)
                    profiler_1.profilerInstance.profileSectionStart('Verify_archiver_receipt');
                if (nestedCounters_1.nestedCountersInstance)
                    nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Verify_archiver_receipt');
                const start_time = process.hrtime();
                // console.log('offloading receipt', txId, timestamp)
                // const result = await offloadReceipt(txId, timestamp, requiredSignatures, receipt)
                let result;
                try {
                    result = await (0, exports.verifyArchiverReceipt)(receipt, requiredSignatures);
                }
                catch (error) {
                    receiptsInValidationMap.delete(txId);
                    if (nestedCounters_1.nestedCountersInstance)
                        nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Invalid_receipt_verification_failed');
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionEnd('Verify_archiver_receipt');
                    continue;
                }
                // console.log('offload receipt result', txId, timestamp, result)
                const end_time = process.hrtime(start_time);
                const time_taken = end_time[0] * 1000 + end_time[1] / 1000000;
                if (time_taken > 100) {
                    console.log(`Time taken for receipt verification in millisecond is: `, txId, timestamp, time_taken);
                }
                if (profiler_1.profilerInstance)
                    profiler_1.profilerInstance.profileSectionEnd('Offload_receipt');
                for (const message of result.failedReasons) {
                    Logger.mainLogger.error(message);
                }
                for (const message of result.nestedCounterMessages) {
                    if (nestedCounters_1.nestedCountersInstance)
                        nestedCounters_1.nestedCountersInstance.countEvent('receipt', message);
                }
                if (result.success === false) {
                    receiptsInValidationMap.delete(txId);
                    if (profiler_1.profilerInstance)
                        profiler_1.profilerInstance.profileSectionEnd('Validate_receipt');
                    continue;
                }
            }
        }
        if (profiler_1.profilerInstance)
            profiler_1.profilerInstance.profileSectionEnd('Validate_receipt');
        // await Receipt.insertReceipt({
        //   ...receipts[i],
        //   receiptId: tx.txId,
        //   timestamp: tx.timestamp,
        // })
        const { afterStates, cycle, tx, appReceiptData, signedReceipt, globalModification } = receipt;
        const sortedVoteOffsets = globalModification
            ? []
            : signedReceipt.voteOffsets.sort();
        const medianOffset = sortedVoteOffsets[Math.floor(sortedVoteOffsets.length / 2)] ?? 0;
        const applyTimestamp = tx.timestamp + medianOffset * 1000;
        if (Config_1.config.VERBOSE)
            console.log('RECEIPT', 'Save', txId, timestamp, senderInfo);
        processedReceiptsMap.set(tx.txId, tx.timestamp);
        receiptsInValidationMap.delete(tx.txId);
        if (missingReceiptsMap.has(tx.txId))
            missingReceiptsMap.delete(tx.txId);
        receipt.beforeStates = globalModification || Config_1.config.storeReceiptBeforeStates ? receipt.beforeStates : []; // Store beforeStates for globalModification tx, or if config.storeReceiptBeforeStates is true
        combineReceipts.push({
            ...receipt,
            receiptId: tx.txId,
            timestamp: tx.timestamp,
            applyTimestamp,
        });
        if (Config_1.config.dataLogWrite && DataLogWriter_1.ReceiptLogWriter)
            DataLogWriter_1.ReceiptLogWriter.writeToLog(`${lib_types_1.Utils.safeStringify({
                ...receipt,
                receiptId: tx.txId,
                timestamp: tx.timestamp,
                applyTimestamp,
            })}\n`);
        txDataList.push({ txId, timestamp });
        // If the receipt is a challenge, then skip updating its accounts data or transaction data
        // if (
        //   config.newPOQReceipt === true &&
        //   appliedReceipt &&
        //   appliedReceipt.confirmOrChallenge &&
        //   appliedReceipt.confirmOrChallenge.message === 'challenge'
        // )
        //   continue
        for (const account of afterStates) {
            const accObj = {
                accountId: account.accountId,
                data: account.data,
                timestamp: account.timestamp,
                hash: account.hash,
                cycleNumber: cycle,
                isGlobal: account.isGlobal || false,
            };
            if (account.timestamp !== account.data['timestamp'])
                Logger.mainLogger.error('Mismatched account timestamp', txId, account.accountId);
            if (account.hash !== account.data['hash'])
                Logger.mainLogger.error('Mismatched account hash', txId, account.accountId);
            const accountExist = await Account.queryAccountByAccountId(account.accountId);
            if (accountExist) {
                if (accObj.timestamp > accountExist.timestamp)
                    await Account.updateAccount(accObj);
            }
            else {
                // await Account.insertAccount(accObj)
                combineAccounts.push(accObj);
            }
            //check global network account updates
            if (accObj.accountId === Config_1.config.globalNetworkAccount) {
                (0, GlobalAccount_1.setGlobalNetworkAccount)(accObj);
            }
            if (accObj.isGlobal) {
                GlobalAccount_1.globalAccountsMap.set(accObj.accountId, {
                    hash: accObj.hash,
                    timestamp: accObj.timestamp,
                });
            }
        }
        // if (receipt) {
        //   const accObj: Account.AccountsCopy = {
        //     accountId: receipt.accountId,
        //     data: receipt.data,
        //     timestamp: receipt.timestamp,
        //     hash: receipt.stateId,
        //     cycleNumber: cycle,
        //   }
        //   const accountExist = await Account.queryAccountByAccountId(
        //     receipt.accountId
        //   )
        //   if (accountExist) {
        //     if (accObj.timestamp > accountExist.timestamp)
        //       await Account.updateAccount(accObj.accountId, accObj)
        //   } else {
        //     // await Account.insertAccount(accObj)
        //     combineAccounts.push(accObj)
        //   }
        // }
        const txObj = {
            txId: tx.txId,
            appReceiptId: appReceiptData ? appReceiptData.accountId : tx.txId,
            timestamp: tx.timestamp,
            cycleNumber: cycle,
            data: appReceiptData ? appReceiptData.data : {},
            originalTxData: tx.originalTxData,
        };
        const processedTx = {
            txId: tx.txId,
            cycle: cycle,
            txTimestamp: tx.timestamp,
            applyTimestamp,
        };
        // await Transaction.insertTransaction(txObj)
        combineTransactions.push(txObj);
        combineProcessedTxs.push(processedTx);
        // Receipts size can be big, better to save per 100
        if (combineReceipts.length >= 100) {
            await Receipt.bulkInsertReceipts(combineReceipts);
            if (State.isActive)
                (0, GossipData_1.sendDataToAdjacentArchivers)(GossipData_1.DataType.RECEIPT, txDataList);
            combineReceipts = [];
            txDataList = [];
        }
        if (combineAccounts.length >= bucketSize) {
            await Account.bulkInsertAccounts(combineAccounts);
            combineAccounts = [];
        }
        if (combineTransactions.length >= bucketSize) {
            await Transaction.bulkInsertTransactions(combineTransactions);
            combineTransactions = [];
        }
        if (combineProcessedTxs.length >= bucketSize) {
            await ProcessedTransaction.bulkInsertProcessedTxs(combineProcessedTxs);
            combineProcessedTxs = [];
        }
    }
    // Receipts size can be big, better to save per 100
    if (combineReceipts.length > 0) {
        await Receipt.bulkInsertReceipts(combineReceipts);
        if (State.isActive)
            (0, GossipData_1.sendDataToAdjacentArchivers)(GossipData_1.DataType.RECEIPT, txDataList);
    }
    if (combineAccounts.length > 0)
        await Account.bulkInsertAccounts(combineAccounts);
    if (combineTransactions.length > 0)
        await Transaction.bulkInsertTransactions(combineTransactions);
    if (combineProcessedTxs.length > 0)
        await ProcessedTransaction.bulkInsertProcessedTxs(combineProcessedTxs);
    // If the archiver is not active, good to clean up the processed receipts map if it exceeds 2000
    if (!State.isActive && processedReceiptsMap.size > 2000)
        processedReceiptsMap.clear();
};
exports.storeReceiptData = storeReceiptData;
const validateCycleData = (cycleRecord) => {
    const err = Utils.validateTypes(cycleRecord, {
        activated: 'a',
        activatedPublicKeys: 'a',
        active: 'n',
        apoptosized: 'a',
        archiverListHash: 's',
        counter: 'n',
        desired: 'n',
        duration: 'n',
        expired: 'n',
        joined: 'a',
        joinedArchivers: 'a',
        joinedConsensors: 'a',
        leavingArchivers: 'a',
        lost: 'a',
        lostSyncing: 'a',
        marker: 's',
        maxSyncTime: 'n',
        mode: 's',
        networkConfigHash: 's',
        networkId: 's',
        nodeListHash: 's',
        previous: 's',
        refreshedArchivers: 'a',
        refreshedConsensors: 'a',
        refuted: 'a',
        removed: 'a',
        returned: 'a',
        standbyAdd: 'a',
        standbyNodeListHash: 's',
        standbyRemove: 'a',
        start: 'n',
        syncing: 'n',
        target: 'n',
        archiversAtShutdown: 'a?',
    });
    if (err) {
        Logger.mainLogger.error('Invalid Cycle Record', err);
        return false;
    }
    const cycleRecordWithoutMarker = { ...cycleRecord };
    delete cycleRecordWithoutMarker.marker;
    if ((0, Cycles_1.computeCycleMarker)(cycleRecordWithoutMarker) !== cycleRecord.marker) {
        Logger.mainLogger.error('Invalid Cycle Record: cycle marker does not match with the computed marker');
        return false;
    }
    return true;
};
exports.validateCycleData = validateCycleData;
const storeCycleData = async (cycles = []) => {
    if (cycles && cycles.length <= 0)
        return;
    const bucketSize = 1000;
    let combineCycles = [];
    for (let i = 0; i < cycles.length; i++) {
        // eslint-disable-next-line security/detect-object-injection
        const cycleRecord = cycles[i];
        const cycleObj = {
            counter: cycleRecord.counter,
            cycleMarker: cycleRecord.marker,
            cycleRecord,
        };
        if (Config_1.config.dataLogWrite && DataLogWriter_1.CycleLogWriter)
            DataLogWriter_1.CycleLogWriter.writeToLog(`${lib_types_1.Utils.safeStringify(cycleObj)}\n`);
        const cycleExist = await (0, cycles_1.queryCycleByMarker)(cycleObj.cycleMarker);
        if (cycleExist) {
            if (lib_types_1.Utils.safeStringify(cycleObj) !== lib_types_1.Utils.safeStringify(cycleExist))
                await (0, cycles_1.updateCycle)(cycleObj.cycleMarker, cycleObj);
        }
        else {
            // await Cycle.insertCycle(cycleObj)
            combineCycles.push(cycleObj);
        }
        if (combineCycles.length >= bucketSize || i === cycles.length - 1) {
            if (combineCycles.length > 0)
                await (0, cycles_1.bulkInsertCycles)(combineCycles);
            combineCycles = [];
        }
    }
};
exports.storeCycleData = storeCycleData;
const storeAccountData = async (restoreData = {}) => {
    Logger.mainLogger.debug(`storeAccountData: ${restoreData.accounts ? restoreData.accounts.length : 0} accounts`);
    Logger.mainLogger.debug(`storeAccountData: ${restoreData.receipts ? restoreData.receipts.length : 0} receipts`);
    const { accounts, receipts } = restoreData;
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionStart('store_account_data');
    exports.storingAccountData = true;
    if (!accounts && !receipts)
        return;
    Logger.mainLogger.debug('Received Accounts Size', accounts ? accounts.length : 0);
    Logger.mainLogger.debug('Received Transactions Size', receipts ? receipts.length : 0);
    // for (let i = 0; i < accounts.length; i++) {
    //   const account = accounts[i]
    //   await Account.insertAccount(account)
    //   // const accountExist = await Account.queryAccountByAccountId(
    //   //   account.accountId
    //   // )
    //   // if (accountExist) {
    //   //   if (account.timestamp > accountExist.timestamp)
    //   //     await Account.updateAccount(account.accountId, account)
    //   // } else {
    //   //   await Account.insertAccount(account)
    //   // }
    // }
    //
    if (accounts && accounts.length > 0) {
        const combineAccounts = [];
        for (const account of accounts) {
            try {
                const calculatedAccountHash = (0, calculateAccountHash_1.accountSpecificHash)(account.data);
                if (calculatedAccountHash !== account.hash) {
                    Logger.mainLogger.error('Invalid account hash', account.accountId, account.hash, calculatedAccountHash);
                    continue;
                }
                combineAccounts.push(account);
            }
            catch (error) {
                Logger.mainLogger.error('Error in calculating genesis account hash', error);
            }
        }
        if (combineAccounts.length > 0)
            await Account.bulkInsertAccounts(accounts);
    }
    if (receipts && receipts.length > 0) {
        Logger.mainLogger.debug('Received receipts Size', receipts.length);
        const combineTransactions = [];
        const combineProcessedTxs = [];
        for (const receipt of receipts) {
            const txObj = {
                txId: receipt.data.txId || receipt.txId,
                appReceiptId: receipt.appReceiptId,
                timestamp: receipt.timestamp,
                cycleNumber: receipt.cycleNumber,
                data: receipt.data,
                originalTxData: {},
            };
            const processedTx = {
                txId: receipt.data.txId || receipt.txId,
                cycle: receipt.cycleNumber,
                txTimestamp: receipt.timestamp,
                applyTimestamp: receipt.timestamp,
            };
            combineTransactions.push(txObj);
            combineProcessedTxs.push(processedTx);
        }
        await Transaction.bulkInsertTransactions(combineTransactions);
        await ProcessedTransaction.bulkInsertProcessedTxs(combineProcessedTxs);
    }
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionEnd('store_account_data');
    Logger.mainLogger.debug('Combined Accounts Data', Data_1.combineAccountsData.accounts.length);
    if (Data_1.combineAccountsData.accounts.length > 0 || Data_1.combineAccountsData.receipts.length > 0) {
        Logger.mainLogger.debug('Found combine accountsData', Data_1.combineAccountsData.accounts.length);
        const accountData = { ...Data_1.combineAccountsData };
        (0, Data_1.clearCombinedAccountsData)();
        (0, exports.storeAccountData)(accountData);
    }
    else {
        exports.storingAccountData = false;
    }
};
exports.storeAccountData = storeAccountData;
const storeOriginalTxData = async (originalTxsData = [], senderInfo = '', saveOnlyGossipData = false) => {
    if (!originalTxsData || !Array.isArray(originalTxsData) || originalTxsData.length <= 0)
        return;
    const bucketSize = 1000;
    let combineOriginalTxsData = [];
    let txDataList = [];
    if (saveOnlyGossipData)
        return;
    for (const originalTxData of originalTxsData) {
        const { txId, timestamp } = originalTxData;
        if (!txId || !timestamp)
            continue;
        try {
            const tx = originalTxData.originalTxData?.tx;
            const { result, reason } = (0, transactionVerification_1.verifyTransaction)(tx);
            if (result !== 'pass') {
                Logger.mainLogger.info(`OriginalTxData verification failed for`, lib_types_1.Utils.safeStringify(originalTxData), '\n with reason ', reason);
                continue;
            }
        }
        catch (error) {
            Logger.mainLogger.error(`Error verifying transaction: ${error.message} where tx was ${lib_types_1.Utils.safeStringify(originalTxData)}`);
            continue;
        }
        if ((processedOriginalTxsMap.has(txId) && processedOriginalTxsMap.get(txId) === timestamp) ||
            (originalTxsInValidationMap.has(txId) && originalTxsInValidationMap.get(txId) === timestamp)) {
            if (Config_1.config.VERBOSE)
                console.log('ORIGINAL_TX_DATA', 'Skip', txId, timestamp, senderInfo);
            continue;
        }
        if (Config_1.config.VERBOSE)
            console.log('ORIGINAL_TX_DATA', 'Validate', txId, timestamp, senderInfo);
        if ((0, exports.validateOriginalTxDataSchema)(originalTxData) === false) {
            Logger.mainLogger.error('Invalid originalTxData: Validation failed', txId);
            originalTxsInValidationMap.delete(txId);
            continue;
        }
        if (Config_1.config.VERBOSE)
            console.log('ORIGINAL_TX_DATA', 'Save', txId, timestamp, senderInfo);
        processedOriginalTxsMap.set(txId, timestamp);
        originalTxsInValidationMap.delete(txId);
        if (missingOriginalTxsMap.has(txId))
            missingOriginalTxsMap.delete(txId);
        if (Config_1.config.dataLogWrite && DataLogWriter_1.OriginalTxDataLogWriter)
            DataLogWriter_1.OriginalTxDataLogWriter.writeToLog(`${lib_types_1.Utils.safeStringify(originalTxData)}\n`);
        combineOriginalTxsData.push(originalTxData);
        txDataList.push({ txId, timestamp });
        if (combineOriginalTxsData.length >= bucketSize) {
            await OriginalTxsData.bulkInsertOriginalTxsData(combineOriginalTxsData);
            if (State.isActive)
                (0, GossipData_1.sendDataToAdjacentArchivers)(GossipData_1.DataType.ORIGINAL_TX_DATA, txDataList);
            combineOriginalTxsData = [];
            txDataList = [];
        }
    }
    if (combineOriginalTxsData.length > 0) {
        await OriginalTxsData.bulkInsertOriginalTxsData(combineOriginalTxsData);
        if (State.isActive)
            (0, GossipData_1.sendDataToAdjacentArchivers)(GossipData_1.DataType.ORIGINAL_TX_DATA, txDataList);
    }
    // If the archiver is not active yet, good to clean up the processed originalTxs map if it exceeds 2000
    if (!State.isActive && processedOriginalTxsMap.size > 2000)
        processedOriginalTxsMap.clear();
};
exports.storeOriginalTxData = storeOriginalTxData;
const validateOriginalTxDataSchema = (originalTxData) => {
    const errors = (0, Helpers_1.verifyPayload)(AJVSchemaEnum_1.AJVSchemaEnum.OriginalTxData, originalTxData);
    if (errors) {
        Logger.mainLogger.error('Invalid originalTxsData', errors, 'where originalTxData was: ', lib_types_1.Utils.safeStringify(originalTxData));
        return false;
    }
    return true;
};
exports.validateOriginalTxDataSchema = validateOriginalTxDataSchema;
const validateGossipData = (data) => {
    let err = Utils.validateTypes(data, {
        dataType: 's',
        data: 'a',
        sign: 'o',
    });
    if (err) {
        Logger.mainLogger.error('Invalid gossip data', data);
        return { success: false, reason: 'Invalid gossip data' + err };
    }
    err = Utils.validateTypes(data.sign, { owner: 's', sig: 's' });
    if (err) {
        Logger.mainLogger.error('Invalid gossip data signature', err);
        return { success: false, reason: 'Invalid gossip data signature' + err };
    }
    if (!State.activeArchivers.some((archiver) => archiver.publicKey === data.sign.owner)) {
        Logger.mainLogger.error('Data sender is not the active archivers');
        return { success: false, error: 'Data sender not the active archivers' };
    }
    if (data.dataType !== GossipData_1.DataType.RECEIPT &&
        data.dataType !== GossipData_1.DataType.ORIGINAL_TX_DATA &&
        data.dataType !== GossipData_1.DataType.CYCLE) {
        Logger.mainLogger.error('Invalid dataType', data);
        return { success: false, error: 'Invalid dataType' };
    }
    if (!Crypto.verify(data)) {
        Logger.mainLogger.error('Invalid signature', data);
        return { success: false, error: 'Invalid signature' };
    }
    return { success: true };
};
exports.validateGossipData = validateGossipData;
const processGossipData = (gossipdata) => {
    const { dataType, data, sign } = gossipdata;
    const senderArchiver = State.activeArchivers.find((archiver) => archiver.publicKey === sign.owner);
    const receivedTimestamp = Date.now();
    if (dataType === GossipData_1.DataType.RECEIPT) {
        for (const { txId, timestamp } of data) {
            if ((processedReceiptsMap.has(txId) && processedReceiptsMap.get(txId) === timestamp) ||
                (receiptsInValidationMap.has(txId) && receiptsInValidationMap.get(txId) === timestamp) ||
                (collectingMissingReceiptsMap.has(txId) && collectingMissingReceiptsMap.get(txId) === timestamp)) {
                // console.log('GOSSIP', 'RECEIPT', 'SKIP', txId, 'sender', sign.owner)
                continue;
            }
            else {
                if (missingReceiptsMap.has(txId)) {
                    if (missingReceiptsMap.get(txId).txTimestamp === timestamp &&
                        !missingReceiptsMap.get(txId).senders.some((sender) => sender === sign.owner))
                        missingReceiptsMap.get(txId).senders.push(sign.owner);
                    else {
                        // Not expected to happen, but log error if it happens <-- could be malicious act of the sender
                        if (missingReceiptsMap.get(txId).txTimestamp !== timestamp)
                            Logger.mainLogger.error(`Received gossip for receipt ${txId} with different timestamp ${timestamp} from archiver ${sign.owner}`);
                        if (missingReceiptsMap.get(txId).senders.some((sender) => sender === sign.owner))
                            Logger.mainLogger.error(`Received gossip for receipt ${txId} from the same sender ${sign.owner}`);
                    }
                }
                else
                    missingReceiptsMap.set(txId, { txTimestamp: timestamp, receivedTimestamp, senders: [sign.owner] });
                // console.log('GOSSIP', 'RECEIPT', 'MISS', txId, 'sender', sign.owner)
            }
        }
    }
    if (dataType === GossipData_1.DataType.ORIGINAL_TX_DATA) {
        for (const { txId, timestamp } of data) {
            if ((processedOriginalTxsMap.has(txId) && processedOriginalTxsMap.get(txId) === timestamp) ||
                (originalTxsInValidationMap.has(txId) && originalTxsInValidationMap.get(txId) === timestamp) ||
                (collectingMissingOriginalTxsMap.has(txId) && collectingMissingOriginalTxsMap.get(txId) === timestamp)) {
                // console.log('GOSSIP', 'ORIGINAL_TX_DATA', 'SKIP', txId, 'sender', sign.owner)
                continue;
            }
            else {
                if (missingOriginalTxsMap.has(txId)) {
                    if (missingOriginalTxsMap.get(txId).txTimestamp === timestamp &&
                        !missingOriginalTxsMap.get(txId).senders.some((sender) => sender === sign.owner))
                        missingOriginalTxsMap.get(txId).senders.push(sign.owner);
                    else {
                        // Not expected to happen, but log error if it happens <-- could be malicious act of the sender
                        if (missingOriginalTxsMap.get(txId).txTimestamp !== timestamp)
                            Logger.mainLogger.error(`Received gossip for originalTxData ${txId} with different timestamp ${timestamp} from archiver ${sign.owner}`);
                        if (missingOriginalTxsMap.get(txId).senders.some((sender) => sender === sign.owner))
                            Logger.mainLogger.error(`Received gossip for originalTxData ${txId} from the same sender ${sign.owner}`);
                    }
                }
                else
                    missingOriginalTxsMap.set(txId, {
                        txTimestamp: timestamp,
                        receivedTimestamp,
                        senders: [sign.owner],
                    });
                // console.log('GOSSIP', 'ORIGINAL_TX_DATA', 'MISS', txId, 'sender', sign.owner)
            }
        }
    }
    if (dataType === GossipData_1.DataType.CYCLE) {
        (0, Data_1.collectCycleData)(data, senderArchiver?.ip + ':' + senderArchiver?.port);
    }
};
exports.processGossipData = processGossipData;
const collectMissingTxDataFromArchivers = async () => {
    const currentTimestamp = Date.now();
    if (missingReceiptsMap.size > 0) {
        const cloneMissingReceiptsMap = new Map();
        for (const [txId, { txTimestamp, receivedTimestamp, senders }] of missingReceiptsMap) {
            if (currentTimestamp - receivedTimestamp > Config_1.config.waitingTimeForMissingTxData) {
                cloneMissingReceiptsMap.set(txId, { txTimestamp, senders });
                collectingMissingReceiptsMap.set(txId, txTimestamp);
                missingReceiptsMap.delete(txId);
            }
        }
        if (cloneMissingReceiptsMap.size > 0)
            Logger.mainLogger.debug('Collecting missing receipts', cloneMissingReceiptsMap.size);
        for (const [txId, { txTimestamp, senders }] of cloneMissingReceiptsMap) {
            (0, exports.collectMissingReceipts)(senders, txId, txTimestamp);
        }
        cloneMissingReceiptsMap.clear();
    }
    if (missingOriginalTxsMap.size > 0) {
        const cloneMissingOriginalTxsMap = new Map();
        for (const [txId, { txTimestamp, receivedTimestamp, senders }] of missingOriginalTxsMap) {
            if (currentTimestamp - receivedTimestamp > Config_1.config.waitingTimeForMissingTxData) {
                cloneMissingOriginalTxsMap.set(txId, { txTimestamp, senders });
                collectingMissingOriginalTxsMap.set(txId, txTimestamp);
                missingOriginalTxsMap.delete(txId);
            }
        }
        if (cloneMissingOriginalTxsMap.size > 0)
            Logger.mainLogger.debug('Collecting missing originalTxsData', cloneMissingOriginalTxsMap.size);
        for (const [txId, { txTimestamp, senders }] of cloneMissingOriginalTxsMap) {
            collectMissingOriginalTxsData(senders, txId, txTimestamp);
        }
        cloneMissingOriginalTxsMap.clear();
    }
};
exports.collectMissingTxDataFromArchivers = collectMissingTxDataFromArchivers;
const collectMissingReceipts = async (senders, txId, txTimestamp) => {
    const txIdList = [[txId, txTimestamp]];
    let foundTxData = false;
    const senderArchivers = State.activeArchivers.filter((archiver) => senders.includes(archiver.publicKey));
    Logger.mainLogger.debug(`Collecting missing receipt for txId ${txId} with timestamp ${txTimestamp} from archivers`, senderArchivers.map((a) => a.ip + ':' + a.port));
    if (nestedCounters_1.nestedCountersInstance)
        nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Collect_missing_receipt');
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionStart('Collect_missing_receipt');
    for (const senderArchiver of senderArchivers) {
        if ((processedReceiptsMap.has(txId) && processedReceiptsMap.get(txId) === txTimestamp) ||
            (receiptsInValidationMap.has(txId) && receiptsInValidationMap.get(txId) === txTimestamp)) {
            foundTxData = true;
            break;
        }
        const receipts = (await (0, exports.queryTxDataFromArchivers)(senderArchiver, GossipData_1.DataType.RECEIPT, txIdList));
        if (receipts && receipts.length > 0) {
            for (const receipt of receipts) {
                const { receiptId, timestamp } = receipt;
                if (txId === receiptId && txTimestamp === timestamp) {
                    (0, exports.storeReceiptData)([receipt], senderArchiver.ip + ':' + senderArchiver.port, true);
                    foundTxData = true;
                }
            }
        }
        if (foundTxData)
            break;
    }
    if (!foundTxData) {
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('receipt', 'Failed to collect missing receipt from archivers');
        Logger.mainLogger.error(`Failed to collect receipt for txId ${txId} with timestamp ${txTimestamp} from archivers ${senders}`);
    }
    collectingMissingReceiptsMap.delete(txId);
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionEnd('Collect_missing_receipt');
};
exports.collectMissingReceipts = collectMissingReceipts;
const collectMissingOriginalTxsData = async (senders, txId, txTimestamp) => {
    const txIdList = [[txId, txTimestamp]];
    let foundTxData = false;
    const senderArchivers = State.activeArchivers.filter((archiver) => senders.includes(archiver.publicKey));
    Logger.mainLogger.debug(`Collecting missing originalTxData for txId ${txId} with timestamp ${txTimestamp} from archivers`, senderArchivers.map((a) => a.ip + ':' + a.port));
    if (nestedCounters_1.nestedCountersInstance)
        nestedCounters_1.nestedCountersInstance.countEvent('originalTxData', 'Collect_missing_originalTxData');
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionStart('Collect_missing_originalTxData');
    for (const senderArchiver of senderArchivers) {
        if ((processedOriginalTxsMap.has(txId) && processedOriginalTxsMap.get(txId) === txTimestamp) ||
            (originalTxsInValidationMap.has(txId) && originalTxsInValidationMap.get(txId) === txTimestamp)) {
            foundTxData = true;
            break;
        }
        const originalTxs = (await (0, exports.queryTxDataFromArchivers)(senderArchiver, GossipData_1.DataType.ORIGINAL_TX_DATA, txIdList));
        if (originalTxs && originalTxs.length > 0) {
            for (const originalTx of originalTxs)
                if (txId === originalTx.txId && txTimestamp === originalTx.timestamp) {
                    (0, exports.storeOriginalTxData)([originalTx], senderArchiver.ip + ':' + senderArchiver.port);
                    foundTxData = true;
                }
        }
        if (foundTxData)
            break;
    }
    if (!foundTxData) {
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('originalTxData', 'Failed to collect_missing_originalTxData');
        Logger.mainLogger.error(`Failed to collect originalTxData for txId ${txId} with timestamp ${txTimestamp} from archivers ${senders}`);
    }
    collectingMissingOriginalTxsMap.delete(txId);
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionEnd('Collect_missing_originalTxData');
};
const queryTxDataFromArchivers = async (archiver, txDataType, txIdList) => {
    let api_route = '';
    if (txDataType === GossipData_1.DataType.RECEIPT) {
        api_route = `receipt`;
    }
    else if (txDataType === GossipData_1.DataType.ORIGINAL_TX_DATA) {
        api_route = `originalTx`;
    }
    const signedData = Crypto.sign({ txIdList, sender: State.getNodeInfo().publicKey });
    const response = (await (0, P2P_1.postJson)(`http://${archiver.ip}:${archiver.port}/${api_route}`, signedData));
    if (response) {
        if (txDataType === GossipData_1.DataType.RECEIPT) {
            const receipts = response.receipts || null;
            if (receipts && receipts.length > -1) {
                return receipts;
            }
        }
        else if (txDataType === GossipData_1.DataType.ORIGINAL_TX_DATA) {
            const originalTxs = response.originalTxs || null;
            if (originalTxs && originalTxs.length > -1) {
                return originalTxs;
            }
        }
    }
    return null;
};
exports.queryTxDataFromArchivers = queryTxDataFromArchivers;
function cleanOldReceiptsMap(timestamp) {
    let savedReceiptsCount = 0;
    for (const [key, value] of processedReceiptsMap) {
        if (value < timestamp) {
            processedReceiptsMap.delete(key);
            savedReceiptsCount++;
        }
    }
    if (savedReceiptsCount > 0)
        Logger.mainLogger.debug(`Clean ${savedReceiptsCount} old receipts from the processed receipts cache on cycle ${(0, Cycles_1.getCurrentCycleCounter)()}`);
}
exports.cleanOldReceiptsMap = cleanOldReceiptsMap;
function cleanOldOriginalTxsMap(timestamp) {
    let savedOriginalTxsCount = 0;
    for (const [key, value] of processedOriginalTxsMap) {
        if (value < timestamp) {
            if (!processedReceiptsMap.has(key))
                Logger.mainLogger.error('The processed receipt is not found for originalTx', key, value);
            processedOriginalTxsMap.delete(key);
            savedOriginalTxsCount++;
        }
    }
    if (savedOriginalTxsCount > 0)
        Logger.mainLogger.debug(`Clean ${savedOriginalTxsCount} old originalTxsData from the processed originalTxsData cache on cycle ${(0, Cycles_1.getCurrentCycleCounter)()}`);
}
exports.cleanOldOriginalTxsMap = cleanOldOriginalTxsMap;
const scheduleMissingTxsDataQuery = () => {
    // Set to collect missing txs data in every 1 second
    setInterval(() => {
        (0, exports.collectMissingTxDataFromArchivers)();
    }, 1000);
};
exports.scheduleMissingTxsDataQuery = scheduleMissingTxsDataQuery;
//# sourceMappingURL=Collector.js.map