"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.sendDataToAdjacentArchivers = exports.getAdjacentLeftAndRightArchivers = exports.DataType = exports.adjacentArchivers = void 0;
const State = require("../State");
const Logger = require("../Logger");
const Crypto = require("../Crypto");
const P2P_1 = require("../P2P");
const Utils = require("../Utils");
const Config_1 = require("../Config");
const nestedCounters_1 = require("../profiler/nestedCounters");
// adjacentArchivers are one archiver from left and one archiver from right of the current archiver
exports.adjacentArchivers = [];
var DataType;
(function (DataType) {
    DataType["RECEIPT"] = "RECEIPT";
    DataType["ORIGINAL_TX_DATA"] = "ORIGINAL_TX_DATA";
    DataType["CYCLE"] = "CYCLE";
})(DataType = exports.DataType || (exports.DataType = {}));
// List of archivers that are not adjacent to the current archiver
const remainingArchivers = [];
const getAdjacentLeftAndRightArchivers = () => {
    if (State.activeArchivers.length <= 1) {
        exports.adjacentArchivers = [];
        return;
    }
    // Treat the archivers list as a circular list and get one left and one right archivers of the current archiver
    const currentArchiverIndex = State.activeArchiversByPublicKeySorted.findIndex((archiver) => archiver.publicKey === State.getNodeInfo().publicKey);
    let leftArchiver = null;
    let rightArchiver = null;
    if (State.activeArchiversByPublicKeySorted.length === 2) {
        if (currentArchiverIndex === 0)
            rightArchiver = State.activeArchiversByPublicKeySorted[1];
        else
            leftArchiver = State.activeArchiversByPublicKeySorted[0];
    }
    else {
        let leftArchiverIndex = currentArchiverIndex - 1;
        let rightArchiverIndex = currentArchiverIndex + 1;
        if (leftArchiverIndex < 0)
            leftArchiverIndex = State.activeArchiversByPublicKeySorted.length - 1;
        if (rightArchiverIndex > State.activeArchiversByPublicKeySorted.length - 1)
            rightArchiverIndex = 0;
        /* eslint-disable security/detect-object-injection */
        leftArchiver = State.activeArchiversByPublicKeySorted[leftArchiverIndex];
        rightArchiver = State.activeArchiversByPublicKeySorted[rightArchiverIndex];
        /* eslint-enable security/detect-object-injection */
    }
    exports.adjacentArchivers.length = 0;
    if (leftArchiver)
        exports.adjacentArchivers.push(leftArchiver);
    if (rightArchiver)
        exports.adjacentArchivers.push(rightArchiver);
    remainingArchivers.length = 0;
    for (const archiver of State.otherArchivers) {
        if (!exports.adjacentArchivers.some((a) => a.publicKey === archiver.publicKey)) {
            remainingArchivers.push(archiver);
        }
    }
};
exports.getAdjacentLeftAndRightArchivers = getAdjacentLeftAndRightArchivers;
async function sendDataToAdjacentArchivers(dataType, data) {
    if (Config_1.config.stopGossipTxData)
        return;
    if (State.otherArchivers.length === 0)
        return;
    const gossipPayload = {
        dataType,
        data,
    };
    const signedDataToSend = Crypto.sign(gossipPayload);
    try {
        const promises = [];
        const archiversToSend = [...exports.adjacentArchivers];
        if (Config_1.config.gossipToMoreArchivers && remainingArchivers.length > 0) {
            const randomArchivers = Utils.getRandomItemFromArr(remainingArchivers, 0, Config_1.config.randomGossipArchiversCount);
            if (randomArchivers.length > 0)
                archiversToSend.push(...randomArchivers);
        }
        if (Config_1.config.VERBOSE)
            Logger.mainLogger.debug(`Sending ${dataType} data to the archivers: ${archiversToSend.map((n) => `${n.ip}:${n.port}`)}`);
        for (const archiver of archiversToSend) {
            const url = `http://${archiver.ip}:${archiver.port}/gossip-data`;
            try {
                const GOSSIP_DATA_TIMEOUT_SECOND = 10; // 10 seconds
                const promise = (0, P2P_1.postJson)(url, signedDataToSend, GOSSIP_DATA_TIMEOUT_SECOND);
                promise.catch((err) => {
                    Logger.mainLogger.error(`Unable to send archiver ${archiver.ip}: ${archiver.port}`, err);
                });
                promises.push(promise);
            }
            catch (e) {
                Logger.mainLogger.error(`Gossip Error to archiver ${archiver.ip}: ${archiver.port}`, e);
            }
        }
        try {
            await Promise.allSettled(promises).then((results) => {
                results.forEach((result) => {
                    if (nestedCounters_1.nestedCountersInstance) {
                        if (result.status === 'fulfilled') {
                            if (result.value !== null)
                                nestedCounters_1.nestedCountersInstance.countEvent('gossip-data', 'success');
                            else
                                nestedCounters_1.nestedCountersInstance.countEvent('gossip-data', 'failure');
                        }
                        else
                            nestedCounters_1.nestedCountersInstance.countEvent('gossip-data', 'failure');
                    }
                });
            });
        }
        catch (err) {
            Logger.mainLogger.error('Gossip Error: ' + err);
            if (nestedCounters_1.nestedCountersInstance)
                nestedCounters_1.nestedCountersInstance.countEvent('gossip-data', 'error 1', err);
        }
    }
    catch (ex) {
        Logger.mainLogger.debug(ex);
        Logger.mainLogger.debug('Fail to gossip');
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('gossip-data', 'error 2', ex);
    }
}
exports.sendDataToAdjacentArchivers = sendDataToAdjacentArchivers;
//# sourceMappingURL=GossipData.js.map