"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.syncGlobalAccount = exports.loadGlobalAccounts = exports.updateGlobalNetworkAccount = exports.setGlobalNetworkAccount = exports.getGlobalNetworkAccount = exports.globalAccountsMap = void 0;
const Crypto = require("./Crypto");
const rfdc = require("rfdc");
const State = require("./State");
const Logger = require("./Logger");
const AccountDB = require("./dbstore/accounts");
const Config_1 = require("./Config");
const P2P_1 = require("./P2P");
const Utils_1 = require("./Utils");
const util_1 = require("util");
const calculateAccountHash_1 = require("./shardeum/calculateAccountHash");
const allowedArchiversManager_1 = require("./shardeum/allowedArchiversManager");
let cachedGlobalNetworkAccount;
let cachedGlobalNetworkAccountHash;
exports.globalAccountsMap = new Map();
const appliedConfigChanges = new Set();
function getGlobalNetworkAccount(hash) {
    if (hash) {
        return cachedGlobalNetworkAccountHash;
    }
    return cachedGlobalNetworkAccount;
}
exports.getGlobalNetworkAccount = getGlobalNetworkAccount;
function setGlobalNetworkAccount(account) {
    cachedGlobalNetworkAccount = rfdc()(account);
    cachedGlobalNetworkAccountHash = account.hash;
    // Get the latest change if any
    allowedArchiversManager_1.allowedArchiversManager.setGlobalAccountConfig(account.data?.listOfChanges?.[account.data?.listOfChanges?.length - 1]?.change?.debug?.multisigKeys, account.data?.listOfChanges?.[account.data?.listOfChanges?.length - 1]?.change?.debug?.minSigRequiredForArchiverWhitelist);
}
exports.setGlobalNetworkAccount = setGlobalNetworkAccount;
const updateGlobalNetworkAccount = async (cycleNumber) => {
    if (!cachedGlobalNetworkAccountHash)
        return;
    const networkAccount = rfdc()(cachedGlobalNetworkAccount);
    const changes = networkAccount.data.listOfChanges;
    if (!changes || !Array.isArray(changes)) {
        return;
    }
    const activeConfigChanges = new Set();
    for (const change of changes) {
        // skip future changes
        if (change.cycle > cycleNumber) {
            continue;
        }
        const changeHash = Crypto.hashObj(change);
        // skip handled changes
        if (appliedConfigChanges.has(changeHash)) {
            activeConfigChanges.add(changeHash);
            continue;
        }
        // apply this change
        appliedConfigChanges.add(changeHash);
        activeConfigChanges.add(changeHash);
        const changeObj = change.change;
        const appData = change.appData;
        // If there is initShutdown change, if the latest cycle is greater than the cycle of the change, then skip it
        if (changeObj['p2p'] && changeObj['p2p']['initShutdown'] && change.cycle !== cycleNumber)
            continue;
        const newChanges = pruneNetworkChangeQueue(changes, cycleNumber);
        networkAccount.data.listOfChanges = newChanges;
        // https://github.com/shardeum/shardeum/blob/c449ecd21391747c5b7173da3a74415da2acb0be/src/index.ts#L6958
        // Increase the timestamp by 1 second
        // networkAccount.data.timestamp += 1000
        if (appData) {
            updateNetworkChangeQueue(networkAccount.data, appData);
            console.dir(networkAccount.data, { depth: null });
            // https://github.com/shardeum/shardeum/blob/c449ecd21391747c5b7173da3a74415da2acb0be/src/index.ts#L6889
            // Increase the timestamp by 1 second
            // networkAccount.data.timestamp += 1000
        }
        networkAccount.hash = (0, calculateAccountHash_1.accountSpecificHash)(networkAccount.data);
        networkAccount.timestamp = networkAccount.data.timestamp;
        Logger.mainLogger.debug('updateGlobalNetworkAccount', networkAccount);
        await AccountDB.updateAccount(networkAccount);
        setGlobalNetworkAccount(networkAccount);
    }
    if (activeConfigChanges.size > 0) {
        // clear the entries from appliedConfigChanges that are no longer in the changes list
        for (const changeHash of appliedConfigChanges) {
            if (!activeConfigChanges.has(changeHash)) {
                appliedConfigChanges.delete(changeHash);
            }
        }
    }
};
exports.updateGlobalNetworkAccount = updateGlobalNetworkAccount;
const generatePathKeys = (obj, prefix = '') => {
    /* eslint-disable security/detect-object-injection */
    let paths = [];
    // Loop over each key in the object
    for (const key of Object.keys(obj)) {
        // If the value corresponding to this key is an object (and not an array or null),
        // then recurse into it.
        if (obj[key] !== null && typeof obj[key] === 'object' && !Array.isArray(obj[key])) {
            paths = paths.concat(generatePathKeys(obj[key], prefix + key + '.'));
        }
        else {
            // Otherwise, just append this key to the path.
            paths.push(prefix + key);
        }
    }
    return paths;
    /* eslint-enable security/detect-object-injection */
};
const pruneNetworkChangeQueue = (changes, currentCycle) => {
    const configsMap = new Map();
    const keepAliveCount = Config_1.config.configChangeMaxChangesToKeep;
    for (let i = changes.length - 1; i >= 0; i--) {
        const thisChange = changes[i];
        let keepAlive = false;
        let appConfigs = [];
        if (thisChange.appData) {
            appConfigs = generatePathKeys(thisChange.appData, 'appdata.');
        }
        const shardusConfigs = generatePathKeys(thisChange.change);
        const allConfigs = appConfigs.concat(shardusConfigs);
        for (const config of allConfigs) {
            if (!configsMap.has(config)) {
                configsMap.set(config, 1);
                keepAlive = true;
            }
            else if (configsMap.get(config) < keepAliveCount) {
                configsMap.set(config, configsMap.get(config) + 1);
                keepAlive = true;
            }
        }
        if (currentCycle - thisChange.cycle <= Config_1.config.configChangeMaxCyclesToKeep) {
            keepAlive = true;
        }
        if (keepAlive == false) {
            changes.splice(i, 1);
        }
    }
    return changes;
};
const updateNetworkChangeQueue = (data, appData) => {
    if ('current' in data)
        patchAndUpdate(data.current, appData);
};
const patchAndUpdate = (existingObject, changeObj, parentPath = '') => {
    /* eslint-disable security/detect-object-injection */
    for (const [key, value] of Object.entries(changeObj)) {
        if (existingObject[key] != null) {
            if (typeof value === 'object') {
                patchAndUpdate(existingObject[key], value, parentPath === '' ? key : parentPath + '.' + key);
            }
            else {
                existingObject[key] = value;
            }
        }
    }
};
const loadGlobalAccounts = async () => {
    const sql = `SELECT * FROM accounts WHERE isGlobal=1`;
    const values = [];
    const accounts = await AccountDB.fetchAccountsBySqlQuery(sql, values);
    for (const account of accounts) {
        exports.globalAccountsMap.set(account.accountId, { hash: account.hash, timestamp: account.timestamp });
        if (account.accountId === Config_1.config.globalNetworkAccount) {
            setGlobalNetworkAccount(account);
        }
    }
};
exports.loadGlobalAccounts = loadGlobalAccounts;
const syncGlobalAccount = async (retry = 5) => {
    while (retry > 0) {
        try {
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            const queryFn = async (node) => {
                return await (0, P2P_1.postJson)(`http://${node.ip}:${node.port}/get_globalaccountreport_archiver`, Crypto.sign({}));
            };
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            const equalFn = (info1, info2) => {
                const cm1 = (0, Utils_1.deepCopy)(info1);
                const cm2 = (0, Utils_1.deepCopy)(info2);
                delete cm1.sign;
                delete cm2.sign;
                const equivalent = (0, util_1.isDeepStrictEqual)(cm1, cm2);
                return equivalent;
            };
            const globalAccsResponse = await (0, Utils_1.robustQuery)(State.otherArchivers, queryFn, equalFn, 3, true);
            Logger.mainLogger.debug('syncGlobalAccount() - globalAccsResponse', globalAccsResponse);
            if (!globalAccsResponse) {
                Logger.mainLogger.warn('() - robustResponse is null');
                throw new Error('() - robustResponse is null');
            }
            const { value: { accounts }, } = globalAccsResponse;
            for (const { id, hash, timestamp } of accounts) {
                exports.globalAccountsMap.set(id, { hash, timestamp });
            }
            if (exports.globalAccountsMap.has(Config_1.config.globalNetworkAccount)) {
                const savedNetworkAccount = await AccountDB.queryAccountByAccountId(Config_1.config.globalNetworkAccount);
                if (savedNetworkAccount &&
                    savedNetworkAccount.hash === exports.globalAccountsMap.get(Config_1.config.globalNetworkAccount).hash) {
                    setGlobalNetworkAccount(savedNetworkAccount);
                    return;
                }
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                const queryFn = async (node) => {
                    return await (0, P2P_1.getJson)(`http://${node.ip}:${node.port}/get-network-account?hash=false`);
                };
                const networkAccResponse = await (0, Utils_1.robustQuery)(State.otherArchivers, queryFn, equalFn, 3, true);
                Logger.mainLogger.debug('syncGlobalAccount() - networkAccResponse', networkAccResponse);
                if (!networkAccResponse) {
                    Logger.mainLogger.warn('get-network-account() - robustResponse is null');
                    throw new Error('get-network-account() - robustResponse is null');
                }
                const { value: { networkAccount }, } = networkAccResponse;
                if (networkAccount) {
                    setGlobalNetworkAccount(networkAccount);
                }
            }
            return;
        }
        catch (e) {
            Logger.mainLogger.error('Error in syncGlobalAccount()', e);
            retry--;
        }
    }
};
exports.syncGlobalAccount = syncGlobalAccount;
//# sourceMappingURL=GlobalAccount.js.map