"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.queryTransactionsBetweenCycles = exports.queryTransactionCountBetweenCycles = exports.queryTransactionCount = exports.queryTransactions = exports.queryLatestTransactions = exports.queryTransactionByAccountId = exports.queryTransactionByTxId = exports.bulkInsertTransactions = exports.insertTransaction = void 0;
// import { Signature } from 'shardus-crypto-types'
const db = require("./sqlite3storage");
const _1 = require(".");
const Logger = require("../Logger");
const Config_1 = require("../Config");
const serialization_1 = require("../utils/serialization");
async function insertTransaction(transaction) {
    try {
        // Define the table columns based on schema
        const columns = ['txId', 'appReceiptId', 'timestamp', 'cycleNumber', 'data', 'originalTxData'];
        // Construct the SQL query with placeholders
        const placeholders = `(${columns.map(() => '?').join(', ')})`;
        const sql = `INSERT OR REPLACE INTO transactions (${columns.join(', ')}) VALUES ${placeholders}`;
        // Map the `transaction` object to match the columns
        const values = columns.map((column) => typeof transaction[column] === 'object'
            ? (0, serialization_1.SerializeToJsonString)(transaction[column]) // Serialize objects to JSON
            : transaction[column]);
        // Execute the query directly
        await db.run(_1.transactionDatabase, sql, values);
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Successfully inserted Transaction', transaction.txId);
        }
    }
    catch (err) {
        Logger.mainLogger.error(err);
        Logger.mainLogger.error('Unable to insert Transaction or it is already stored in the database', transaction.txId);
    }
}
exports.insertTransaction = insertTransaction;
async function bulkInsertTransactions(transactions) {
    try {
        // Define the table columns based on schema
        const columns = ['txId', 'appReceiptId', 'timestamp', 'cycleNumber', 'data', 'originalTxData'];
        // Construct the SQL query for bulk insertion with all placeholders
        const placeholders = transactions.map(() => `(${columns.map(() => '?').join(', ')})`).join(', ');
        const sql = `INSERT OR REPLACE INTO transactions (${columns.join(', ')}) VALUES ${placeholders}`;
        // Flatten the `transactions` array into a single list of values
        const values = transactions.flatMap((transaction) => columns.map((column) => typeof transaction[column] === 'object'
            ? (0, serialization_1.SerializeToJsonString)(transaction[column]) // Serialize objects to JSON
            : transaction[column]));
        // Execute the single query for all transactions
        await db.run(_1.transactionDatabase, sql, values);
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Successfully inserted Transactions', transactions.length);
        }
    }
    catch (err) {
        Logger.mainLogger.error(err);
        Logger.mainLogger.error('Unable to bulk insert Transactions', transactions.length);
    }
}
exports.bulkInsertTransactions = bulkInsertTransactions;
async function queryTransactionByTxId(txId) {
    try {
        const sql = `SELECT * FROM transactions WHERE txId=?`;
        const transaction = (await db.get(_1.transactionDatabase, sql, [txId])); // TODO: confirm structure of object from db
        if (transaction) {
            if (transaction.data)
                transaction.data = (0, serialization_1.DeSerializeFromJsonString)(transaction.data);
            if (transaction.originalTxData)
                transaction.originalTxData = (0, serialization_1.DeSerializeFromJsonString)(transaction.originalTxData);
        }
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Transaction txId', transaction);
        }
        return transaction;
    }
    catch (e) {
        Logger.mainLogger.error(e);
        return null;
    }
}
exports.queryTransactionByTxId = queryTransactionByTxId;
async function queryTransactionByAccountId(accountId) {
    try {
        const sql = `SELECT * FROM transactions WHERE accountId=?`;
        const transaction = (await db.get(_1.transactionDatabase, sql, [accountId])); // TODO: confirm structure of object from db
        if (transaction) {
            if (transaction.data)
                transaction.data = (0, serialization_1.DeSerializeFromJsonString)(transaction.data);
            if (transaction.originalTxData)
                transaction.originalTxData = (0, serialization_1.DeSerializeFromJsonString)(transaction.originalTxData);
        }
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Transaction accountId', transaction);
        }
        return transaction;
    }
    catch (e) {
        Logger.mainLogger.error(e);
        return null;
    }
}
exports.queryTransactionByAccountId = queryTransactionByAccountId;
async function queryLatestTransactions(count) {
    if (!Number.isInteger(count)) {
        Logger.mainLogger.error('queryLatestTransactions - Invalid count value');
        return null;
    }
    try {
        const sql = `SELECT * FROM transactions ORDER BY cycleNumber DESC, timestamp DESC LIMIT ${count ? count : 100}`;
        const transactions = (await db.all(_1.transactionDatabase, sql)); // TODO: confirm structure of object from db
        if (transactions.length > 0) {
            transactions.forEach((transaction) => {
                if (transaction.data)
                    transaction.data = (0, serialization_1.DeSerializeFromJsonString)(transaction.data);
                if (transaction.originalTxData)
                    transaction.originalTxData = (0, serialization_1.DeSerializeFromJsonString)(transaction.originalTxData);
            });
        }
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug('Transaction latest', transactions);
        }
        return transactions;
    }
    catch (e) {
        Logger.mainLogger.error(e);
        return null;
    }
}
exports.queryLatestTransactions = queryLatestTransactions;
async function queryTransactions(skip = 0, limit = 10000) {
    let transactions;
    if (!Number.isInteger(skip) || !Number.isInteger(limit)) {
        Logger.mainLogger.error('queryTransactions - Invalid skip or limit');
        return null;
    }
    try {
        const sql = `SELECT * FROM transactions ORDER BY cycleNumber ASC, timestamp ASC LIMIT ${limit} OFFSET ${skip}`;
        transactions = (await db.all(_1.transactionDatabase, sql)); // TODO: confirm structure of object from db
        if (transactions.length > 0) {
            transactions.forEach((transaction) => {
                if (transaction.data)
                    transaction.data = (0, serialization_1.DeSerializeFromJsonString)(transaction.data);
                if (transaction.originalTxData)
                    transaction.originalTxData = (0, serialization_1.DeSerializeFromJsonString)(transaction.originalTxData);
            });
        }
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('Transaction transactions', transactions ? transactions.length : transactions, 'skip', skip);
    }
    return transactions;
}
exports.queryTransactions = queryTransactions;
async function queryTransactionCount() {
    let transactions;
    try {
        const sql = `SELECT COUNT(*) FROM transactions`;
        transactions = await db.get(_1.transactionDatabase, sql, []);
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('Transaction count', transactions);
    }
    if (transactions)
        transactions = transactions['COUNT(*)'];
    else
        transactions = 0;
    return transactions;
}
exports.queryTransactionCount = queryTransactionCount;
async function queryTransactionCountBetweenCycles(startCycleNumber, endCycleNumber) {
    let transactions;
    try {
        const sql = `SELECT COUNT(*) FROM transactions WHERE cycleNumber BETWEEN ? AND ?`;
        transactions = await db.get(_1.transactionDatabase, sql, [startCycleNumber, endCycleNumber]);
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('Transaction count between cycles', transactions);
    }
    if (transactions)
        transactions = transactions['COUNT(*)'];
    else
        transactions = 0;
    return transactions;
}
exports.queryTransactionCountBetweenCycles = queryTransactionCountBetweenCycles;
async function queryTransactionsBetweenCycles(skip = 0, limit = 10000, startCycleNumber, endCycleNumber) {
    let transactions;
    if (!Number.isInteger(skip) || !Number.isInteger(limit)) {
        Logger.mainLogger.error('queryTransactionsBetweenCycles - Invalid skip or limit value');
        return null;
    }
    try {
        const sql = `SELECT * FROM transactions WHERE cycleNumber BETWEEN ? AND ? ORDER BY cycleNumber ASC, timestamp ASC LIMIT ${limit} OFFSET ${skip}`;
        transactions = (await db.all(_1.transactionDatabase, sql, [
            startCycleNumber,
            endCycleNumber,
        ])); // TODO: confirm structure of object from db
        if (transactions.length > 0) {
            transactions.forEach((transaction) => {
                if (transaction.data)
                    transaction.data = (0, serialization_1.DeSerializeFromJsonString)(transaction.data);
                if (transaction.originalTxData)
                    transaction.originalTxData = (0, serialization_1.DeSerializeFromJsonString)(transaction.originalTxData);
            });
        }
    }
    catch (e) {
        Logger.mainLogger.error(e);
    }
    if (Config_1.config.VERBOSE) {
        Logger.mainLogger.debug('Transaction transactions between cycles', transactions ? transactions.length : transactions, 'skip', skip);
    }
    return transactions;
}
exports.queryTransactionsBetweenCycles = queryTransactionsBetweenCycles;
//# sourceMappingURL=transactions.js.map