"use strict";
/**
 * `queries` submodule. Contains logic pertaining to anything that
 * requires an external node to be queried, including robust queries.
 *
 * Code is duplicated from shardus-global-server.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.getTxListFromNode = exports.getStandbyNodeListFromNode = exports.getArchiverListFromNode = exports.getValidatorListFromNode = exports.getCurrentCycleDataFromNode = exports.robustQueryForTxListHash = exports.robustQueryForStandbyNodeListHash = exports.robustQueryForArchiverListHash = exports.robustQueryForValidatorListHash = exports.robustQueryForCycleRecordHash = void 0;
const neverthrow_1 = require("neverthrow");
const Utils_1 = require("../Utils");
const P2P_1 = require("../P2P");
const MAX_RETRIES = 3;
const REDUNDANCY = 3;
async function throwableQueryFunction(node, endpointName, params) {
    let url = `http://${node.ip}:${node.port}/${endpointName}`;
    if (params) {
        const encodedParams = new URLSearchParams(params).toString();
        url += `?${encodedParams}`;
    }
    const res = await (0, P2P_1.get)(url);
    if (res.ok) {
        return (await res.json());
    }
    else {
        throw new Error(`get failed with status ${res.statusText}`);
    }
}
/**
 * Executes a robust query to a specified endpoint across multiple nodes, providing more fault tolerance.
 *
 * @param {ActiveNode[]} nodes - An array of active nodes to query.
 * @param {string} endpointName - The name of the endpoint to call on the active nodes.
 *
 * The function runs a robust query, logs the query, and ensures the result of the query is robust.
 * It makes an HTTP GET request to the specified endpoint for each node, and if the call fails,
 * an error message is returned.
 *
 * @returns {RobustQueryResultAsync<T>} - A ResultAsync object. On success, it contains a result object with
 * the winning nodes and the returned value. On failure, it contains an Error object. The function is asynchronous
 * and can be awaited.
 */
function makeRobustQueryCall(nodes, endpointName) {
    // query function that makes the endpoint call as specified
    const queryFn = (node) => neverthrow_1.ResultAsync.fromPromise(throwableQueryFunction(node, endpointName, {}), (err) => new Error(`couldn't query ${endpointName}: ${err}`));
    // run the robust query, wrapped in an async Result return the unwrapped result (with `map`) if successful
    const logPrefix = `syncv2-robust-query-${endpointName}`;
    return neverthrow_1.ResultAsync.fromPromise((0, Utils_1.attempt)(() => (0, Utils_1.robustQuery)(nodes, queryFn), {
        maxRetries: MAX_RETRIES,
        logPrefix,
    }), (err) => new Error(`robust query failed for ${endpointName}: ${err}`)).andThen((robustResult) => {
        // ensure the result was robust as well.
        if (!(robustResult.count >= Math.min(REDUNDANCY, nodes.length))) {
            return (0, neverthrow_1.errAsync)(new Error(`result of ${endpointName} wasn't robust`));
        }
        return robustResult.value.map((value) => ({
            winningNodes: robustResult.nodes,
            value,
        }));
    });
}
/**
 * Performs a simple fetch operation from a given node to a specified endpoint. The operation is retried up to
 * MAX_RETRIES times in case of failure.
 *
 * @param {ActiveNode} node - An active node to fetch data from.
 * @param {string} endpointName - The name of the endpoint to fetch data from.
 *
 * The function attempts to make an HTTP GET request to the specified endpoint. If the call fails,
 * it retries the operation and returns an error message.
 *
 * @returns {ResultAsync<T, Error>} - A ResultAsync object. On success, it will contain the data fetched, and
 * on error, it will contain an Error object. The function is asynchronous and can be awaited.
 */
function attemptSimpleFetch(node, endpointName, params) {
    return neverthrow_1.ResultAsync.fromPromise((0, Utils_1.attempt)(async () => throwableQueryFunction(node, endpointName, params), {
        maxRetries: MAX_RETRIES,
        logPrefix: `syncv2-simple-fetch-${endpointName}`,
    }), (err) => new Error(`simple fetch failed for ${endpointName}: ${err}`));
}
/** Executes a robust query to retrieve the cycle marker from the network. */
function robustQueryForCycleRecordHash(nodes) {
    return makeRobustQueryCall(nodes, 'current-cycle-hash');
}
exports.robustQueryForCycleRecordHash = robustQueryForCycleRecordHash;
/** Executes a robust query to retrieve the validator list hash and next cycle timestamp from the network. */
function robustQueryForValidatorListHash(nodes) {
    return makeRobustQueryCall(nodes, 'validator-list-hash');
}
exports.robustQueryForValidatorListHash = robustQueryForValidatorListHash;
/** Executes a robust query to retrieve the archiver list hash from the network. */
function robustQueryForArchiverListHash(nodes) {
    return makeRobustQueryCall(nodes, 'archiver-list-hash');
}
exports.robustQueryForArchiverListHash = robustQueryForArchiverListHash;
/** Executes a robust query to retrieve the standby list hash from the network. */
function robustQueryForStandbyNodeListHash(nodes) {
    return makeRobustQueryCall(nodes, 'standby-list-hash');
}
exports.robustQueryForStandbyNodeListHash = robustQueryForStandbyNodeListHash;
/** Executes a robust query to retrieve the txList hash from the network. */
function robustQueryForTxListHash(nodes) {
    return makeRobustQueryCall(nodes, 'tx-list-hash');
}
exports.robustQueryForTxListHash = robustQueryForTxListHash;
/** Retrives the entire last cycle from the node. */
function getCurrentCycleDataFromNode(node, expectedMarker) {
    return attemptSimpleFetch(node, 'cycle-by-marker', {
        marker: expectedMarker,
    });
}
exports.getCurrentCycleDataFromNode = getCurrentCycleDataFromNode;
/** Gets the full validator list from the specified node. */
function getValidatorListFromNode(node, expectedHash) {
    console.log(`getting validator list from ${node.ip}:${node.port} with hash ${expectedHash}`);
    return attemptSimpleFetch(node, 'validator-list', {
        hash: expectedHash,
    });
}
exports.getValidatorListFromNode = getValidatorListFromNode;
/** Gets the full node list from the specified archiver. */
function getArchiverListFromNode(node, expectedHash) {
    console.log(`getting archiver list from ${node.ip}:${node.port} with hash ${expectedHash}`);
    return attemptSimpleFetch(node, 'archiver-list', {
        hash: expectedHash,
    });
}
exports.getArchiverListFromNode = getArchiverListFromNode;
/** Gets the full standby list from the specified node. */
function getStandbyNodeListFromNode(node, expectedHash) {
    return attemptSimpleFetch(node, 'standby-list', {
        hash: expectedHash,
    });
}
exports.getStandbyNodeListFromNode = getStandbyNodeListFromNode;
/** Gets the full tx list from the specified node */
function getTxListFromNode(node, expectedHash) {
    return attemptSimpleFetch(node, 'tx-list', {
        hash: expectedHash,
    });
}
exports.getTxListFromNode = getTxListFromNode;
//# sourceMappingURL=queries.js.map